"""Group of methods for node operation."""

import logging
import pathlib as pl

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class NodeGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj

    def gen_kes_key_pair(
        self, node_name: str, destination_dir: itp.FileType = "."
    ) -> structs.KeyPair:
        """Generate a key pair for a node KES operational key.

        Args:
            node_name: A name of the node the key pair is generated for.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            structs.KeyPair: A data container containing the key pair.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        vkey = destination_dir / f"{node_name}_kes.vkey"
        skey = destination_dir / f"{node_name}_kes.skey"
        clusterlib_helpers._check_files_exist(vkey, skey, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                "node",
                "key-gen-KES",
                "--verification-key-file",
                str(vkey),
                "--signing-key-file",
                str(skey),
            ]
        )

        helpers._check_outfiles(vkey, skey)
        return structs.KeyPair(vkey, skey)

    def gen_vrf_key_pair(
        self, node_name: str, destination_dir: itp.FileType = "."
    ) -> structs.KeyPair:
        """Generate a key pair for a node VRF operational key.

        Args:
            node_name: A name of the node the key pair is generated for.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            structs.KeyPair: A data container containing the key pair.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        vkey = destination_dir / f"{node_name}_vrf.vkey"
        skey = destination_dir / f"{node_name}_vrf.skey"
        clusterlib_helpers._check_files_exist(vkey, skey, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                "node",
                "key-gen-VRF",
                "--verification-key-file",
                str(vkey),
                "--signing-key-file",
                str(skey),
            ]
        )

        helpers._check_outfiles(vkey, skey)
        return structs.KeyPair(vkey, skey)

    def gen_cold_key_pair_and_counter(
        self, node_name: str, destination_dir: itp.FileType = "."
    ) -> structs.ColdKeyPair:
        """Generate a key pair for operator's offline key and a new certificate issue counter.

        Args:
            node_name: A name of the node the key pair and the counter is generated for.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            structs.ColdKeyPair: A data container containing the key pair and the counter.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        vkey = destination_dir / f"{node_name}_cold.vkey"
        skey = destination_dir / f"{node_name}_cold.skey"
        counter = destination_dir / f"{node_name}_cold.counter"
        clusterlib_helpers._check_files_exist(
            vkey, skey, counter, clusterlib_obj=self._clusterlib_obj
        )

        self._clusterlib_obj.cli(
            [
                "node",
                "key-gen",
                "--cold-verification-key-file",
                str(vkey),
                "--cold-signing-key-file",
                str(skey),
                "--operational-certificate-issue-counter-file",
                str(counter),
            ]
        )

        helpers._check_outfiles(vkey, skey, counter)
        return structs.ColdKeyPair(vkey, skey, counter)

    def gen_node_operational_cert(
        self,
        node_name: str,
        kes_vkey_file: itp.FileType,
        cold_skey_file: itp.FileType,
        cold_counter_file: itp.FileType,
        kes_period: int | None = None,
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Generate a node operational certificate.

        This certificate is used when starting the node and not submitted through a transaction.

        Args:
            node_name: A name of the node the certificate is generated for.
            kes_vkey_file: A path to pool KES vkey file.
            cold_skey_file: A path to pool cold skey file.
            cold_counter_file: A path to pool cold counter file.
            kes_period: A start KES period. The current KES period is used when not specified.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to the generated certificate.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{node_name}.opcert"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        kes_period = (
            kes_period if kes_period is not None else self._clusterlib_obj.g_query.get_kes_period()
        )

        self._clusterlib_obj.cli(
            [
                "node",
                "issue-op-cert",
                "--kes-verification-key-file",
                str(kes_vkey_file),
                "--cold-signing-key-file",
                str(cold_skey_file),
                "--operational-certificate-issue-counter",
                str(cold_counter_file),
                "--kes-period",
                str(kes_period),
                "--out-file",
                str(out_file),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: clusterlib_obj={id(self._clusterlib_obj)}>"



"""Legacy top-level module.

Import everything that used to be available here for backwards compatibility.
"""

# flake8: noqa
from cardano_clusterlib.clusterlib_klass import ClusterLib
from cardano_clusterlib.consts import CommandEras
from cardano_clusterlib.consts import DEFAULT_COIN
from cardano_clusterlib.consts import Eras
from cardano_clusterlib.consts import MAINNET_MAGIC
from cardano_clusterlib.consts import MultiSigTypeArgs
from cardano_clusterlib.consts import MultiSlotTypeArgs
from cardano_clusterlib.consts import ScriptTypes
from cardano_clusterlib.consts import Votes
from cardano_clusterlib.exceptions import CLIError
from cardano_clusterlib.helpers import get_rand_str
from cardano_clusterlib.helpers import read_address_from_file
from cardano_clusterlib.structs import ActionConstitution
from cardano_clusterlib.structs import ActionHardfork
from cardano_clusterlib.structs import ActionInfo
from cardano_clusterlib.structs import ActionNoConfidence
from cardano_clusterlib.structs import ActionPParamsUpdate
from cardano_clusterlib.structs import ActionTreasuryWithdrawal
from cardano_clusterlib.structs import ActionUpdateCommittee
from cardano_clusterlib.structs import AddressInfo
from cardano_clusterlib.structs import AddressRecord
from cardano_clusterlib.structs import CCMember
from cardano_clusterlib.structs import CLIOut
from cardano_clusterlib.structs import ColdKeyPair
from cardano_clusterlib.structs import ComplexCert
from cardano_clusterlib.structs import ComplexProposal
from cardano_clusterlib.structs import DataForBuild
from cardano_clusterlib.structs import GenesisKeys
from cardano_clusterlib.structs import KeyPair
from cardano_clusterlib.structs import LeadershipSchedule
from cardano_clusterlib.structs import Mint
from cardano_clusterlib.structs import OptionalMint
from cardano_clusterlib.structs import OptionalScriptCerts
from cardano_clusterlib.structs import OptionalScriptProposals
from cardano_clusterlib.structs import OptionalScriptTxIn
from cardano_clusterlib.structs import OptionalScriptVotes
from cardano_clusterlib.structs import OptionalScriptWithdrawals
from cardano_clusterlib.structs import OptionalTxOuts
from cardano_clusterlib.structs import OptionalUTXOData
from cardano_clusterlib.structs import PoolCreationOutput
from cardano_clusterlib.structs import PoolData
from cardano_clusterlib.structs import PoolParamsTop
from cardano_clusterlib.structs import PoolUser
from cardano_clusterlib.structs import ScriptTxIn
from cardano_clusterlib.structs import ScriptVote
from cardano_clusterlib.structs import ScriptWithdrawal
from cardano_clusterlib.structs import StakeAddrInfo
from cardano_clusterlib.structs import TxFiles
from cardano_clusterlib.structs import TxOut
from cardano_clusterlib.structs import TxRawOutput
from cardano_clusterlib.structs import UTXOData
from cardano_clusterlib.structs import Value
from cardano_clusterlib.structs import VoteCC
from cardano_clusterlib.structs import VoteDrep
from cardano_clusterlib.structs import VoteSPO
from cardano_clusterlib.txtools import calculate_utxos_balance
from cardano_clusterlib.txtools import collect_data_for_build
from cardano_clusterlib.txtools import filter_utxo_with_highest_amount
from cardano_clusterlib.txtools import filter_utxos
from cardano_clusterlib.types import FileType



"""Group of methods for Conway governance vote commands."""

import json
import logging
import pathlib as pl
import typing as tp

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import consts
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class GovVoteGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj
        self._group_args = ("governance", "vote")

    def _get_vote_args(
        self,
        vote: consts.Votes,
    ) -> list[str]:
        if vote == consts.Votes.YES:
            vote_args = ["--yes"]
        elif vote == consts.Votes.NO:
            vote_args = ["--no"]
        elif vote == consts.Votes.ABSTAIN:
            vote_args = ["--abstain"]
        else:
            msg = "No vote was specified."
            raise AssertionError(msg)

        return vote_args

    def _get_gov_action_args(
        self,
        action_txid: str,
        action_ix: int,
    ) -> list[str]:
        gov_action_args = [
            "--governance-action-tx-id",
            str(action_txid),
            "--governance-action-index",
            str(action_ix),
        ]
        return gov_action_args

    def _get_anchor_args(
        self,
        anchor_url: str = "",
        anchor_data_hash: str = "",
    ) -> list[str]:
        anchor_args = []
        if anchor_url:
            if not anchor_data_hash:
                msg = "Anchor data hash is required when anchor URL is specified."
                raise AssertionError(msg)
            anchor_args = [
                "--anchor-url",
                str(anchor_url),
                "--anchor-data-hash",
                str(anchor_data_hash),
            ]
        return anchor_args

    def create_committee(
        self,
        vote_name: str,
        action_txid: str,
        action_ix: int,
        vote: consts.Votes,
        cc_hot_vkey: str = "",
        cc_hot_vkey_file: itp.FileType | None = None,
        cc_hot_key_hash: str = "",
        cc_hot_script_hash: str = "",
        anchor_url: str = "",
        anchor_data_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> structs.VoteCC:
        """Create a governance action vote for a commitee member."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{vote_name}_cc.vote"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        vote_args = self._get_vote_args(vote=vote)
        gov_action_args = self._get_gov_action_args(action_txid=action_txid, action_ix=action_ix)
        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url, anchor_data_hash=anchor_data_hash
        )

        if cc_hot_vkey:
            cred_args = ["--cc-hot-verification-key", cc_hot_vkey]
        elif cc_hot_vkey_file:
            cred_args = ["--cc-hot-verification-key-file", str(cc_hot_vkey_file)]
        elif cc_hot_key_hash:
            cred_args = ["--cc-hot-key-hash", cc_hot_key_hash]
        elif cc_hot_script_hash:
            cred_args = ["--cc-hot-script-hash", cc_hot_script_hash]
        else:
            msg = "No CC key or script hash was specified."
            raise AssertionError(msg)

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create",
                *vote_args,
                *gov_action_args,
                *cred_args,
                *anchor_args,
                "--out-file",
                str(out_file),
            ]
        )
        helpers._check_outfiles(out_file)

        vote_cc = structs.VoteCC(
            action_txid=action_txid,
            action_ix=action_ix,
            vote=vote,
            vote_file=out_file,
            cc_hot_vkey=cc_hot_vkey,
            cc_hot_vkey_file=helpers._maybe_path(cc_hot_vkey_file),
            cc_hot_key_hash=cc_hot_key_hash,
            cc_hot_script_hash=cc_hot_script_hash,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )
        return vote_cc

    def create_drep(
        self,
        vote_name: str,
        action_txid: str,
        action_ix: int,
        vote: consts.Votes,
        drep_vkey: str = "",
        drep_vkey_file: itp.FileType | None = None,
        drep_key_hash: str = "",
        drep_script_hash: str = "",
        anchor_url: str = "",
        anchor_data_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> structs.VoteDrep:
        """Create a governance action vote for a DRep."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{vote_name}_drep.vote"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        vote_args = self._get_vote_args(vote=vote)
        gov_action_args = self._get_gov_action_args(action_txid=action_txid, action_ix=action_ix)
        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url, anchor_data_hash=anchor_data_hash
        )

        if drep_vkey:
            cred_args = ["--drep-verification-key", drep_vkey]
        elif drep_vkey_file:
            cred_args = ["--drep-verification-key-file", str(drep_vkey_file)]
        elif drep_key_hash:
            cred_args = ["--drep-key-hash", drep_key_hash]
        elif drep_script_hash:
            cred_args = ["--drep-script-hash", drep_script_hash]
        else:
            msg = "No DRep key or script hash was specified."
            raise AssertionError(msg)

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create",
                *vote_args,
                *gov_action_args,
                *cred_args,
                *anchor_args,
                "--out-file",
                str(out_file),
            ]
        )
        helpers._check_outfiles(out_file)

        vote_drep = structs.VoteDrep(
            action_txid=action_txid,
            action_ix=action_ix,
            vote=vote,
            vote_file=out_file,
            drep_vkey=drep_vkey,
            drep_vkey_file=helpers._maybe_path(drep_vkey_file),
            drep_key_hash=drep_key_hash,
            drep_script_hash=drep_script_hash,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )
        return vote_drep

    def create_spo(
        self,
        vote_name: str,
        action_txid: str,
        action_ix: int,
        vote: consts.Votes,
        stake_pool_vkey: str = "",
        cold_vkey_file: itp.FileType | None = None,
        stake_pool_id: str = "",
        anchor_url: str = "",
        anchor_data_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> structs.VoteSPO:
        """Create a governance action vote for an SPO."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{vote_name}_spo.vote"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        vote_args = self._get_vote_args(vote=vote)
        gov_action_args = self._get_gov_action_args(action_txid=action_txid, action_ix=action_ix)
        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url, anchor_data_hash=anchor_data_hash
        )

        if stake_pool_vkey:
            key_args = ["--stake-pool-verification-key", stake_pool_vkey]
        elif cold_vkey_file:
            key_args = ["--cold-verification-key-file", str(cold_vkey_file)]
        elif stake_pool_id:
            key_args = ["--stake-pool-id", stake_pool_id]
        else:
            msg = "No key was specified."
            raise AssertionError(msg)

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create",
                *vote_args,
                *gov_action_args,
                *key_args,
                *anchor_args,
                "--out-file",
                str(out_file),
            ]
        )
        helpers._check_outfiles(out_file)

        vote_drep = structs.VoteSPO(
            action_txid=action_txid,
            action_ix=action_ix,
            vote=vote,
            stake_pool_vkey=stake_pool_vkey,
            cold_vkey_file=helpers._maybe_path(cold_vkey_file),
            stake_pool_id=stake_pool_id,
            vote_file=out_file,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )
        return vote_drep

    def view(self, vote_file: itp.FileType) -> dict[str, tp.Any]:
        """View a governance action vote."""
        vote_file = pl.Path(vote_file).expanduser()
        clusterlib_helpers._check_files_exist(vote_file, clusterlib_obj=self._clusterlib_obj)

        stdout = self._clusterlib_obj.cli(
            [
                *self._group_args,
                "view",
                "--vote-file",
                str(vote_file),
            ]
        ).stdout.strip()
        stdout_dec = stdout.decode("utf-8") if stdout else ""

        out: dict[str, tp.Any] = json.loads(stdout_dec)
        return out



"""Group of methods for working with transactions."""

import itertools
import json
import logging
import pathlib as pl
import typing as tp

from packaging import version

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import consts
from cardano_clusterlib import exceptions
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import txtools
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class TransactionGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj
        self.min_fee = self._clusterlib_obj.genesis["protocolParams"]["minFeeB"]
        self._has_debug_prop: bool | None = None

    @property
    def _has_debug(self) -> bool:
        """Check if cardano-cli has a `debug transaction` group."""
        if self._has_debug_prop is not None:
            return self._has_debug_prop

        err = ""
        try:
            self._clusterlib_obj.cli(
                ["cardano-cli", "debug", "transaction"], add_default_args=False
            )
        except exceptions.CLIError as excp:
            err = str(excp)

        self._has_debug_prop = "Usage:" in err
        return self._has_debug_prop

    def calculate_tx_ttl(self) -> int:
        """Calculate ttl for a transaction."""
        return self._clusterlib_obj.g_query.get_slot_no() + self._clusterlib_obj.ttl_length

    def get_txid(self, tx_body_file: itp.FileType = "", tx_file: itp.FileType = "") -> str:
        """Return the transaction identifier.

        Args:
            tx_body_file: A path to the transaction body file (JSON TxBody - optional).
            tx_file: A path to the signed transaction file (JSON Tx - optional).

        Returns:
            str: A transaction ID.
        """
        cli_args = []

        # The `--output-text` option is available since 10.3.0.0 and is the default
        # until about 10.8.0.0. Then the JSON format became the default.
        # Adding the option for cli >= 10.7.0.0, because the option was already available.
        if self._clusterlib_obj.cli_version >= version.parse("10.7.0.0"):
            cli_args.append("--output-text")

        if tx_body_file:
            cli_args.extend(["--tx-body-file", str(tx_body_file)])
        elif tx_file:
            cli_args.extend(["--tx-file", str(tx_file)])
        else:
            msg = "Either `tx_body_file` or `tx_file` is needed."
            raise AssertionError(msg)

        return (
            self._clusterlib_obj.cli(["transaction", "txid", *cli_args])
            .stdout.rstrip()
            .decode("ascii")
        )

    def view_tx(self, tx_body_file: itp.FileType = "", tx_file: itp.FileType = "") -> str:
        """View a transaction.

        Args:
            tx_body_file: A path to the transaction body file (JSON TxBody - optional).
            tx_file: A path to the signed transaction file (JSON Tx - optional).

        Returns:
            str: A transaction.
        """
        cli_args = ["transaction", "view", "--output-json"]

        if tx_body_file:
            cli_args.extend(["--tx-body-file", str(tx_body_file)])
        elif tx_file:
            cli_args.extend(["--tx-file", str(tx_file)])
        else:
            msg = "Either `tx_body_file` or `tx_file` is needed."
            raise AssertionError(msg)

        if self._has_debug:
            cli_args = ["cardano-cli", "debug", *cli_args]

        return (
            self._clusterlib_obj.cli(cli_args, add_default_args=not self._has_debug)
            .stdout.rstrip()
            .decode("utf-8")
        )

    def view_tx_dict(self, tx_body_file: itp.FileType = "", tx_file: itp.FileType = "") -> dict:
        """Get a dict from transaction view.

        Args:
            tx_body_file: A path to the transaction body file (JSON TxBody - optional).
            tx_file: A path to the signed transaction file (JSON Tx - optional).

        Returns:
            dict: A transaction.
        """
        out: dict = json.loads(self.view_tx(tx_body_file=tx_body_file, tx_file=tx_file))
        return out

    def get_hash_script_data(
        self,
        script_data_file: itp.FileType | None = None,
        script_data_cbor_file: itp.FileType | None = None,
        script_data_value: str = "",
    ) -> str:
        """Return the hash of script data.

        Args:
            script_data_file: A path to the JSON file containing the script data (optional).
            script_data_cbor_file: A path to the CBOR file containing the script data (optional).
            script_data_value: A value (in JSON syntax) for the script data (optional).

        Returns:
            str: A hash of script data.
        """
        if script_data_file:
            cli_args = ["--script-data-file", str(script_data_file)]
        elif script_data_cbor_file:
            cli_args = ["--script-data-cbor-file", str(script_data_cbor_file)]
        elif script_data_value:
            cli_args = ["--script-data-value", str(script_data_value)]
        else:
            msg = (
                "Either `script_data_file`, `script_data_cbor_file` or `script_data_value` "
                "is needed."
            )
            raise AssertionError(msg)

        return (
            self._clusterlib_obj.cli(["transaction", "hash-script-data", *cli_args])
            .stdout.rstrip()
            .decode("ascii")
        )

    def get_tx_deposit(self, tx_files: structs.TxFiles) -> int:
        """Get deposit amount for a transaction (based on certificates used for the TX).

        Args:
            tx_files: A `structs.TxFiles` data container containing files needed
                for the transaction.

        Returns:
            int: A total deposit amount needed for the transaction.
        """
        if not tx_files.certificate_files:
            return 0

        pparams = self._clusterlib_obj.g_query.get_protocol_params()
        key_deposit = self._clusterlib_obj.g_query.get_address_deposit(pparams=pparams)
        pool_deposit = self._clusterlib_obj.g_query.get_pool_deposit(pparams=pparams)
        drep_deposit = self._clusterlib_obj.g_query.get_drep_deposit(pparams=pparams)
        action_deposit = self._clusterlib_obj.g_query.get_gov_action_deposit(pparams=pparams)

        deposit = 0

        for cert in tx_files.certificate_files:
            with open(cert, encoding="utf-8") as in_json:
                content = json.load(in_json)
            description = content.get("description", "")
            if (
                "Stake Address Registration" in description
                or "Stake address registration and" in description
            ):
                deposit += key_deposit
            elif "Stake Address Deregistration" in description:
                deposit -= key_deposit
            elif "Stake Pool Registration" in description:
                deposit += pool_deposit
            elif "DRep Key Registration" in description:
                deposit += drep_deposit
            elif "DRep Retirement" in description:
                deposit -= drep_deposit

        for prop in tx_files.proposal_files:
            with open(prop, encoding="utf-8") as in_json:
                content = json.load(in_json)
            ptype = content.get("type", "")
            if "Governance proposal" in ptype:
                deposit += action_deposit

        return deposit

    def build_raw_tx_bare(  # noqa: C901
        self,
        out_file: itp.FileType,
        txouts: list[structs.TxOut],
        tx_files: structs.TxFiles,
        fee: int,
        txins: structs.OptionalUTXOData = (),
        readonly_reference_txins: structs.OptionalUTXOData = (),
        script_txins: structs.OptionalScriptTxIn = (),
        return_collateral_txouts: structs.OptionalTxOuts = (),
        total_collateral_amount: int | None = None,
        mint: structs.OptionalMint = (),
        complex_certs: structs.OptionalScriptCerts = (),
        complex_proposals: structs.OptionalScriptProposals = (),
        required_signers: itp.OptionalFiles = (),
        required_signer_hashes: tp.Optional[list[str]] = None,
        ttl: int | None = None,
        withdrawals: structs.OptionalTxOuts = (),
        script_withdrawals: structs.OptionalScriptWithdrawals = (),
        script_votes: structs.OptionalScriptVotes = (),
        invalid_hereafter: int | None = None,
        invalid_before: int | None = None,
        current_treasury_value: int | None = None,
        treasury_donation: int | None = None,
        script_valid: bool = True,
        join_txouts: bool = True,
    ) -> structs.TxRawOutput:
        """Build a raw transaction.

        Args:
            out_file: An output file.
            txouts: A list (iterable) of `TxOuts`, specifying transaction outputs.
            tx_files: A `structs.TxFiles` data container containing files needed
                for the transaction.
            fee: A fee amount.
            txins: An iterable of `structs.UTXOData`, specifying input UTxOs (optional).
            readonly_reference_txins: An iterable of `structs.UTXOData`, specifying input
                UTxOs to be referenced and used as readonly (optional).
            script_txins: An iterable of `ScriptTxIn`, specifying input script UTxOs (optional).
            return_collateral_txouts: A list (iterable) of `TxOuts`, specifying transaction outputs
                for excess collateral (optional).
            total_collateral_amount: An integer indicating the total amount of collateral
                (optional).
            mint: An iterable of `Mint`, specifying script minting data (optional).
            complex_certs: An iterable of `ComplexCert`, specifying certificates script data
                (optional).
            complex_proposals: An iterable of `ComplexProposal`, specifying proposal script data
                (optional).
            required_signers: An iterable of filepaths of the signing keys whose signatures
                are required (optional).
            required_signer_hashes: A list of hashes of the signing keys whose signatures
                are required (optional).
            ttl: A last block when the transaction is still valid
                (deprecated in favor of `invalid_hereafter`, optional).
            withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals (optional).
            script_withdrawals: An iterable of `ScriptWithdrawal`, specifying withdrawal script
                data (optional).
            script_votes: An iterable of `ScriptVote`, specifying vote script data (optional).
            invalid_hereafter: A last block when the transaction is still valid (optional).
            invalid_before: A first block when the transaction is valid (optional).
            current_treasury_value: The current treasury value (optional).
            treasury_donation: A donation to the treasury to perform (optional).
            script_valid: A bool indicating that the script is valid (True by default).
            join_txouts: A bool indicating whether to aggregate transaction outputs
                by payment address (True by default).

        Returns:
            structs.TxRawOutput: A data container with transaction output details.
        """
        if (treasury_donation is not None) != (current_treasury_value is not None):
            msg = (
                "Both `treasury_donation` and `current_treasury_value` must be specified together."
            )
            raise AssertionError(msg)

        if tx_files.certificate_files and complex_certs:
            LOGGER.warning(
                "Mixing `tx_files.certificate_files` and `complex_certs`, "
                "certs may come in unexpected order."
            )

        if tx_files.proposal_files and complex_proposals:
            LOGGER.warning(
                "Mixing `tx_files.proposal_files` and `complex_proposals`, "
                "proposals may come in unexpected order."
            )

        out_file = pl.Path(out_file)

        withdrawals, script_withdrawals, __ = txtools._get_withdrawals(
            clusterlib_obj=self._clusterlib_obj,
            withdrawals=withdrawals,
            script_withdrawals=script_withdrawals,
        )

        required_signer_hashes = required_signer_hashes or []

        txout_args, processed_txouts, txouts_count = txtools._process_txouts(
            txouts=txouts, join_txouts=join_txouts
        )

        txin_strings = txtools._get_txin_strings(txins=txins, script_txins=script_txins)

        withdrawal_strings = {f"{x.address}+{x.amount}" for x in withdrawals}

        mint_txouts = list(itertools.chain.from_iterable(m.txouts for m in mint))

        misc_args = []

        if invalid_before is not None:
            misc_args.extend(["--invalid-before", str(invalid_before)])
        if invalid_hereafter is not None:
            misc_args.extend(["--invalid-hereafter", str(invalid_hereafter)])
        elif ttl is not None:
            # `--ttl` and `--invalid-hereafter` are the same
            misc_args.extend(["--ttl", str(ttl)])

        if current_treasury_value is not None:
            misc_args.extend(["--current-treasury-value", str(current_treasury_value)])
        if treasury_donation is not None:
            misc_args.extend(["--treasury-donation", str(treasury_donation)])

        if not script_valid:
            misc_args.append("--script-invalid")

        # Only single `--mint` argument is allowed, let's aggregate all the outputs
        mint_records = [f"{m.amount} {m.coin}" for m in mint_txouts]
        misc_args.extend(["--mint", "+".join(mint_records)] if mint_records else [])

        for txin in readonly_reference_txins:
            misc_args.extend(["--read-only-tx-in-reference", f"{txin.utxo_hash}#{txin.utxo_ix}"])

        grouped_args = txtools._get_script_args(
            script_txins=script_txins,
            mint=mint,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            script_withdrawals=script_withdrawals,
            script_votes=script_votes,
            with_execution_units=True,
        )

        grouped_args_str = " ".join(grouped_args)
        pparams_for_txins = grouped_args and (
            "-datum-" in grouped_args_str or "-redeemer-" in grouped_args_str
        )
        # TODO: see https://github.com/input-output-hk/cardano-node/issues/4058
        pparams_for_txouts = "datum-embed-" in " ".join(txout_args)
        if pparams_for_txins or pparams_for_txouts:
            self._clusterlib_obj.create_pparams_file()
            grouped_args.extend(
                [
                    "--protocol-params-file",
                    str(self._clusterlib_obj.pparams_file),
                ]
            )

        if total_collateral_amount:
            misc_args.extend(["--tx-total-collateral", str(total_collateral_amount)])

        if tx_files.metadata_json_files and tx_files.metadata_json_detailed_schema:
            misc_args.append("--json-metadata-detailed-schema")

        proposal_file_argname = txtools.get_proposal_file_argname(
            era_in_use=self._clusterlib_obj.era_in_use
        )

        cli_args = [
            "transaction",
            "build-raw",
            "--fee",
            str(fee),
            "--out-file",
            str(out_file),
            *grouped_args,
            *helpers._prepend_flag("--tx-in", txin_strings),
            *txout_args,
            *helpers._prepend_flag("--required-signer", required_signers),
            *helpers._prepend_flag("--required-signer-hash", required_signer_hashes),
            *helpers._prepend_flag("--certificate-file", tx_files.certificate_files),
            *helpers._prepend_flag(proposal_file_argname, tx_files.proposal_files),
            *helpers._prepend_flag("--vote-file", tx_files.vote_files),
            *helpers._prepend_flag("--auxiliary-script-file", tx_files.auxiliary_script_files),
            *helpers._prepend_flag("--metadata-json-file", tx_files.metadata_json_files),
            *helpers._prepend_flag("--metadata-cbor-file", tx_files.metadata_cbor_files),
            *helpers._prepend_flag("--withdrawal", withdrawal_strings),
            *txtools._get_return_collateral_txout_args(txouts=return_collateral_txouts),
            *misc_args,
        ]

        self._clusterlib_obj.cli(cli_args)

        return structs.TxRawOutput(
            txins=list(txins),
            txouts_count=txouts_count,
            txouts=processed_txouts,
            tx_files=tx_files,
            out_file=out_file,
            fee=fee,
            build_args=cli_args,
            era=self._clusterlib_obj.era_in_use,
            script_txins=script_txins,
            script_withdrawals=script_withdrawals,
            script_votes=script_votes,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            mint=mint,
            invalid_hereafter=invalid_hereafter or ttl,
            invalid_before=invalid_before,
            current_treasury_value=current_treasury_value,
            treasury_donation=treasury_donation,
            withdrawals=withdrawals,
            return_collateral_txouts=return_collateral_txouts,
            total_collateral_amount=total_collateral_amount,
            readonly_reference_txins=readonly_reference_txins,
            script_valid=script_valid,
            required_signers=required_signers,
            required_signer_hashes=required_signer_hashes,
            combined_reference_txins=txtools._get_reference_txins(
                readonly_reference_txins=readonly_reference_txins,
                script_txins=script_txins,
                mint=mint,
                complex_certs=complex_certs,
                script_withdrawals=script_withdrawals,
            ),
        )

    def build_raw_tx(
        self,
        src_address: str,
        tx_name: str,
        txins: structs.OptionalUTXOData = (),
        txouts: structs.OptionalTxOuts = (),
        readonly_reference_txins: structs.OptionalUTXOData = (),
        script_txins: structs.OptionalScriptTxIn = (),
        return_collateral_txouts: structs.OptionalTxOuts = (),
        total_collateral_amount: int | None = None,
        mint: structs.OptionalMint = (),
        tx_files: structs.TxFiles | None = None,
        complex_certs: structs.OptionalScriptCerts = (),
        complex_proposals: structs.OptionalScriptProposals = (),
        fee: int = 0,
        required_signers: itp.OptionalFiles = (),
        required_signer_hashes: tp.Optional[list[str]] = None,
        ttl: int | None = None,
        withdrawals: structs.OptionalTxOuts = (),
        script_withdrawals: structs.OptionalScriptWithdrawals = (),
        script_votes: structs.OptionalScriptVotes = (),
        deposit: int | None = None,
        current_treasury_value: int | None = None,
        treasury_donation: int | None = None,
        invalid_hereafter: int | None = None,
        invalid_before: int | None = None,
        script_valid: bool = True,
        src_addr_utxos: list[structs.UTXOData] | None = None,
        join_txouts: bool = True,
        destination_dir: itp.FileType = ".",
    ) -> structs.TxRawOutput:
        """Balance inputs and outputs and build a raw transaction.

        Args:
            src_address: An address used for fee and inputs (if inputs not specified by `txins`).
            tx_name: A name of the transaction.
            txins: An iterable of `structs.UTXOData`, specifying input UTxOs (optional).
            txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
            readonly_reference_txins: An iterable of `structs.UTXOData`, specifying input
                UTxOs to be referenced and used as readonly (optional).
            script_txins: An iterable of `ScriptTxIn`, specifying input script UTxOs (optional).
            return_collateral_txouts: A list (iterable) of `TxOuts`, specifying transaction outputs
                for excess collateral (optional).
            total_collateral_amount: An integer indicating the total amount of collateral
                (optional).
            mint: An iterable of `Mint`, specifying script minting data (optional).
            tx_files: A `structs.TxFiles` data container containing files needed for the transaction
                (optional).
            complex_certs: An iterable of `ComplexCert`, specifying certificates script data
                (optional).
            complex_proposals: An iterable of `ComplexProposal`, specifying proposals script data
                (optional).
            fee: A fee amount (optional).
            required_signers: An iterable of filepaths of the signing keys whose signatures
                are required (optional).
            required_signer_hashes: A list of hashes of the signing keys whose signatures
                are required (optional).
            ttl: A last block when the transaction is still valid
                (deprecated in favor of `invalid_hereafter`, optional).
            withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals (optional).
            script_withdrawals: An iterable of `ScriptWithdrawal`, specifying withdrawal script
                data (optional).
            script_votes: An iterable of `ScriptVote`, specifying vote script data (optional).
            deposit: A deposit amount needed by the transaction (optional).
            current_treasury_value: The current treasury value (optional).
            treasury_donation: A donation to the treasury to perform (optional).
            invalid_hereafter: A last block when the transaction is still valid (optional).
            invalid_before: A first block when the transaction is valid (optional).
            script_valid: A bool indicating that the script is valid (True by default).
            src_addr_utxos: A list of UTxOs for the source address (optional).
            join_txouts: A bool indicating whether to aggregate transaction outputs
                by payment address (True by default).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            structs.TxRawOutput: A data container with transaction output details.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{tx_name}_tx.body"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        tx_files = tx_files or structs.TxFiles()

        collected_data = txtools.collect_data_for_build(
            clusterlib_obj=self._clusterlib_obj,
            src_address=src_address,
            txins=txins,
            txouts=txouts,
            script_txins=script_txins,
            mint=mint,
            tx_files=tx_files,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            fee=fee,
            withdrawals=withdrawals,
            script_withdrawals=script_withdrawals,
            deposit=deposit,
            treasury_donation=treasury_donation,
            src_addr_utxos=src_addr_utxos,
            skip_asset_balancing=False,
        )

        if (
            ttl is None
            and invalid_hereafter is None
            and (self._clusterlib_obj.era_in_use == consts.Eras.SHELLEY.name.lower())
        ):
            invalid_hereafter = self.calculate_tx_ttl()

        tx_raw_output = self.build_raw_tx_bare(
            out_file=out_file,
            txouts=collected_data.txouts,
            tx_files=tx_files,
            fee=fee,
            txins=collected_data.txins,
            readonly_reference_txins=readonly_reference_txins,
            script_txins=script_txins,
            return_collateral_txouts=return_collateral_txouts,
            total_collateral_amount=total_collateral_amount,
            mint=mint,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            required_signers=required_signers,
            required_signer_hashes=required_signer_hashes,
            withdrawals=collected_data.withdrawals,
            script_withdrawals=collected_data.script_withdrawals,
            script_votes=script_votes,
            invalid_hereafter=invalid_hereafter or ttl,
            invalid_before=invalid_before,
            current_treasury_value=current_treasury_value,
            treasury_donation=treasury_donation,
            script_valid=script_valid,
            join_txouts=join_txouts,
        )

        helpers._check_outfiles(out_file)
        return tx_raw_output

    def estimate_fee(
        self,
        txbody_file: itp.FileType,
        txin_count: int,
        txout_count: int,
        witness_count: int,
        byron_witness_count: int = 0,
        reference_script_size: int = 0,
    ) -> int:
        """Estimate the minimum fee for a transaction.

        Args:
            txbody_file: A path to file with transaction body.
            txin_count: A number of transaction inputs.
            txout_count: A number of transaction outputs.
            witness_count: A number of witnesses.
            byron_witness_count: A number of Byron witnesses (optional).
            reference_script_size: A size in bytes of transaction reference scripts (optional).

        Returns:
            int: An estimated fee.
        """
        cli_args = []
        if self._clusterlib_obj.cli_version >= version.parse("8.22.0.0"):
            cli_args = ["--reference-script-size", str(reference_script_size)]

        self._clusterlib_obj.create_pparams_file()
        stdout = self._clusterlib_obj.cli(
            [
                "transaction",
                "calculate-min-fee",
                "--output-text",
                *self._clusterlib_obj.magic_args,
                "--protocol-params-file",
                str(self._clusterlib_obj.pparams_file),
                "--tx-in-count",
                str(txin_count),
                "--tx-out-count",
                str(txout_count),
                "--byron-witness-count",
                str(byron_witness_count),
                "--witness-count",
                str(witness_count),
                "--tx-body-file",
                str(txbody_file),
                *cli_args,
            ]
        ).stdout
        fee, *__ = stdout.decode().split()
        return int(fee)

    def calculate_tx_fee(
        self,
        src_address: str,
        tx_name: str,
        dst_addresses: tp.Optional[list[str]] = None,
        txins: structs.OptionalUTXOData = (),
        txouts: structs.OptionalTxOuts = (),
        readonly_reference_txins: structs.OptionalUTXOData = (),
        script_txins: structs.OptionalScriptTxIn = (),
        return_collateral_txouts: structs.OptionalTxOuts = (),
        total_collateral_amount: int | None = None,
        mint: structs.OptionalMint = (),
        tx_files: structs.TxFiles | None = None,
        complex_certs: structs.OptionalScriptCerts = (),
        complex_proposals: structs.OptionalScriptProposals = (),
        required_signers: itp.OptionalFiles = (),
        required_signer_hashes: tp.Optional[list[str]] = None,
        ttl: int | None = None,
        withdrawals: structs.OptionalTxOuts = (),
        script_withdrawals: structs.OptionalScriptWithdrawals = (),
        script_votes: structs.OptionalScriptVotes = (),
        deposit: int | None = None,
        current_treasury_value: int | None = None,
        treasury_donation: int | None = None,
        invalid_hereafter: int | None = None,
        invalid_before: int | None = None,
        src_addr_utxos: list[structs.UTXOData] | None = None,
        witness_count_add: int = 0,
        join_txouts: bool = True,
        destination_dir: itp.FileType = ".",
    ) -> int:
        """Build "dummy" transaction and calculate (estimate) its fee.

        Args:
            src_address: An address used for fee and inputs (if inputs not specified by `txins`).
            tx_name: A name of the transaction.
            dst_addresses: A list of destination addresses (optional)
            txins: An iterable of `structs.UTXOData`, specifying input UTxOs (optional).
            txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
            readonly_reference_txins: An iterable of `structs.UTXOData`, specifying input
                UTxOs to be referenced and used as readonly (optional).
            script_txins: An iterable of `ScriptTxIn`, specifying input script UTxOs (optional).
            return_collateral_txouts: A list (iterable) of `TxOuts`, specifying transaction outputs
                for excess collateral (optional).
            total_collateral_amount: An integer indicating the total amount of collateral
                (optional).
            mint: An iterable of `Mint`, specifying script minting data (optional).
            tx_files: A `structs.TxFiles` data container containing files needed for the transaction
                (optional).
            complex_certs: An iterable of `ComplexCert`, specifying certificates script data
                (optional).
            complex_proposals: An iterable of `ComplexProposal`, specifying proposal script data
                (optional).
            required_signers: An iterable of filepaths of the signing keys whose signatures
                are required (optional).
            required_signer_hashes: A list of hashes of the signing keys whose signatures
                are required (optional).
            ttl: A last block when the transaction is still valid
                (deprecated in favor of `invalid_hereafter`, optional).
            withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals (optional).
            script_withdrawals: An iterable of `ScriptWithdrawal`, specifying withdrawal script
                data (optional).
            script_votes: An iterable of `ScriptVote`, specifying vote script data (optional).
            deposit: A deposit amount needed by the transaction (optional).
            current_treasury_value: The current treasury value (optional).
            treasury_donation: A donation to the treasury to perform (optional).
            invalid_hereafter: A last block when the transaction is still valid (optional).
            invalid_before: A first block when the transaction is valid (optional).
            src_addr_utxos: A list of UTxOs for the source address (optional).
            witness_count_add: A number of witnesses to add - workaround to make the fee
                calculation more precise.
            join_txouts: A bool indicating whether to aggregate transaction outputs
                by payment address (True by default).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            int: An estimated fee.
        """
        tx_files = tx_files or structs.TxFiles()
        tx_name = f"{tx_name}_estimate"

        if dst_addresses and txouts:
            LOGGER.warning(
                "The value of `dst_addresses` is ignored when value for `txouts` is available."
            )

        txouts_filled = txouts or [
            structs.TxOut(address=r, amount=1) for r in (dst_addresses or ())
        ]

        tx_raw_output = self.build_raw_tx(
            src_address=src_address,
            tx_name=tx_name,
            txins=txins,
            txouts=txouts_filled,
            readonly_reference_txins=readonly_reference_txins,
            script_txins=script_txins,
            return_collateral_txouts=return_collateral_txouts,
            total_collateral_amount=total_collateral_amount,
            mint=mint,
            tx_files=tx_files,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            required_signers=required_signers,
            required_signer_hashes=required_signer_hashes,
            fee=self.min_fee,
            withdrawals=withdrawals,
            script_withdrawals=script_withdrawals,
            script_votes=script_votes,
            invalid_hereafter=invalid_hereafter or ttl,
            invalid_before=invalid_before,
            src_addr_utxos=src_addr_utxos,
            deposit=deposit,
            current_treasury_value=current_treasury_value,
            treasury_donation=treasury_donation,
            join_txouts=join_txouts,
            destination_dir=destination_dir,
        )

        fee = self.estimate_fee(
            txbody_file=tx_raw_output.out_file,
            # +1 as possibly one more input will be needed for the fee amount
            txin_count=len(tx_raw_output.txins) + 1,
            txout_count=len(tx_raw_output.txouts),
            witness_count=len(tx_files.signing_key_files) + witness_count_add,
        )

        return fee

    def calculate_min_req_utxo(
        self,
        txouts: list[structs.TxOut],
    ) -> structs.Value:
        """Calculate the minimum required UTxO for a single transaction output.

        Args:
            txouts: A list of `TxOut` records that correspond to a single transaction output (UTxO).

        Returns:
            structs.Value: A data container describing the value.
        """
        if not txouts:
            msg = "No txout was specified."
            raise AssertionError(msg)

        txout_args, __, txouts_count = txtools._join_txouts(txouts=txouts)

        if txouts_count > 1:
            msg = (
                "Accepts `TxOuts` only for a single transaction txout "
                "(same address, datum, script)."
            )
            raise AssertionError(msg)

        self._clusterlib_obj.create_pparams_file()
        stdout = self._clusterlib_obj.cli(
            [
                "transaction",
                "calculate-min-required-utxo",
                "--protocol-params-file",
                str(self._clusterlib_obj.pparams_file),
                *txout_args,
            ]
        ).stdout
        coin, value = stdout.decode().split()
        return structs.Value(value=int(value), coin=coin)

    def build_tx(  # noqa: C901
        self,
        src_address: str,
        tx_name: str,
        txins: structs.OptionalUTXOData = (),
        txouts: structs.OptionalTxOuts = (),
        readonly_reference_txins: structs.OptionalUTXOData = (),
        script_txins: structs.OptionalScriptTxIn = (),
        return_collateral_txouts: structs.OptionalTxOuts = (),
        total_collateral_amount: int | None = None,
        mint: structs.OptionalMint = (),
        tx_files: structs.TxFiles | None = None,
        complex_certs: structs.OptionalScriptCerts = (),
        complex_proposals: structs.OptionalScriptProposals = (),
        change_address: str = "",
        fee_buffer: int | None = None,
        required_signers: itp.OptionalFiles = (),
        required_signer_hashes: tp.Optional[list[str]] = None,
        withdrawals: structs.OptionalTxOuts = (),
        script_withdrawals: structs.OptionalScriptWithdrawals = (),
        script_votes: structs.OptionalScriptVotes = (),
        deposit: int | None = None,
        treasury_donation: int | None = None,
        invalid_hereafter: int | None = None,
        invalid_before: int | None = None,
        witness_override: int | None = None,
        script_valid: bool = True,
        calc_script_cost_file: itp.FileType | None = None,
        join_txouts: bool = True,
        destination_dir: itp.FileType = ".",
        skip_asset_balancing: bool = True,
    ) -> structs.TxRawOutput:
        """Build a transaction.

        Args:
            src_address: An address used for fee and inputs (if inputs not specified by `txins`).
            tx_name: A name of the transaction.
            txins: An iterable of `structs.UTXOData`, specifying input UTxOs (optional).
            txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
            readonly_reference_txins: An iterable of `structs.UTXOData`, specifying input
                UTxOs to be referenced and used as readonly (optional).
            script_txins: An iterable of `ScriptTxIn`, specifying input script UTxOs (optional).
            return_collateral_txouts: A list (iterable) of `TxOuts`, specifying transaction outputs
                for excess collateral (optional).
            total_collateral_amount: An integer indicating the total amount of collateral
                (optional).
            mint: An iterable of `Mint`, specifying script minting data (optional).
            tx_files: A `structs.TxFiles` data container containing files needed for the transaction
                (optional).
            complex_certs: An iterable of `ComplexCert`, specifying certificates script data
                (optional).
            complex_proposals: An iterable of `ComplexProposal`, specifying proposal script data
                (optional).
            change_address: A string with address where ADA in excess of the transaction fee
                will go to (`src_address` by default).
            fee_buffer: A buffer for fee amount (optional).
            required_signers: An iterable of filepaths of the signing keys whose signatures
                are required (optional).
            required_signer_hashes: A list of hashes of the signing keys whose signatures
                are required (optional).
            withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals (optional).
            script_withdrawals: An iterable of `ScriptWithdrawal`, specifying withdrawal script
                data (optional).
            script_votes: An iterable of `ScriptVote`, specifying vote script data (optional).
            deposit: A deposit amount needed by the transaction (optional).
            treasury_donation: A donation to the treasury to perform (optional).
            invalid_hereafter: A last block when the transaction is still valid (optional).
            invalid_before: A first block when the transaction is valid (optional).
            witness_override: An integer indicating real number of witnesses. Can be used to fix
                fee calculation (optional).
            script_valid: A bool indicating that the script is valid (True by default).
            calc_script_cost_file: A path for output of the Plutus script cost information
                (optional).
            join_txouts: A bool indicating whether to aggregate transaction outputs
                by payment address (True by default).
            destination_dir: A path to directory for storing artifacts (optional).
            skip_asset_balancing: A bool indicating if assets balancing should be skipped
                (`build` command balance the assets automatically in newer versions).

        Returns:
            structs.TxRawOutput: A data container with transaction output details.
        """
        max_txout = [o for o in txouts if o.amount == -1 and o.coin in ("", consts.DEFAULT_COIN)]
        if max_txout:
            if change_address:
                msg = "Cannot use '-1' amount and change address at the same time."
                raise AssertionError(msg)
            change_address = max_txout[0].address
        else:
            change_address = change_address or src_address

        tx_files = tx_files or structs.TxFiles()
        if tx_files.certificate_files and complex_certs:
            LOGGER.warning(
                "Mixing `tx_files.certificate_files` and `complex_certs`, "
                "certs may come in unexpected order."
            )
        if tx_files.proposal_files and complex_proposals:
            LOGGER.warning(
                "Mixing `tx_files.proposal_files` and `complex_proposals`, "
                "proposals may come in unexpected order."
            )

        destination_dir = pl.Path(destination_dir).expanduser()

        out_file = destination_dir / f"{tx_name}_tx.body"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        collected_data = txtools.collect_data_for_build(
            clusterlib_obj=self._clusterlib_obj,
            src_address=src_address,
            txins=txins,
            txouts=txouts,
            script_txins=script_txins,
            mint=mint,
            tx_files=tx_files,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            fee=fee_buffer or 0,
            withdrawals=withdrawals,
            script_withdrawals=script_withdrawals,
            deposit=deposit,
            treasury_donation=treasury_donation,
            skip_asset_balancing=skip_asset_balancing,
        )

        required_signer_hashes = required_signer_hashes or []

        txout_args, processed_txouts, txouts_count = txtools._process_txouts(
            txouts=collected_data.txouts, join_txouts=join_txouts
        )

        txin_strings = txtools._get_txin_strings(
            txins=collected_data.txins, script_txins=script_txins
        )

        withdrawal_strings = [f"{x.address}+{x.amount}" for x in collected_data.withdrawals]

        mint_txouts = list(itertools.chain.from_iterable(m.txouts for m in mint))

        misc_args = []

        if invalid_before is not None:
            misc_args.extend(["--invalid-before", str(invalid_before)])
        if invalid_hereafter is not None:
            misc_args.extend(["--invalid-hereafter", str(invalid_hereafter)])

        if treasury_donation is not None:
            misc_args.extend(["--treasury-donation", str(treasury_donation)])

        if not script_valid:
            misc_args.append("--script-invalid")

        # There's allowed just single `--mint` argument, let's aggregate all the outputs
        mint_records = [f"{m.amount} {m.coin}" for m in mint_txouts]
        misc_args.extend(["--mint", "+".join(mint_records)] if mint_records else [])

        for txin in readonly_reference_txins:
            misc_args.extend(["--read-only-tx-in-reference", f"{txin.utxo_hash}#{txin.utxo_ix}"])

        grouped_args = txtools._get_script_args(
            script_txins=script_txins,
            mint=mint,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            script_withdrawals=collected_data.script_withdrawals,
            script_votes=script_votes,
            with_execution_units=False,
        )

        if witness_override is not None:
            misc_args.extend(["--witness-override", str(witness_override)])

        if total_collateral_amount:
            misc_args.extend(["--tx-total-collateral", str(total_collateral_amount)])

        if calc_script_cost_file:
            misc_args.extend(["--calculate-plutus-script-cost", str(calc_script_cost_file)])
            out_file = pl.Path(calc_script_cost_file)
        else:
            misc_args.extend(["--out-file", str(out_file)])

        if tx_files.metadata_json_files and tx_files.metadata_json_detailed_schema:
            misc_args.append("--json-metadata-detailed-schema")

        proposal_file_argname = txtools.get_proposal_file_argname(
            era_in_use=self._clusterlib_obj.era_in_use
        )

        cli_args = [
            "transaction",
            "build",
            *grouped_args,
            *helpers._prepend_flag("--tx-in", txin_strings),
            *txout_args,
            *helpers._prepend_flag("--required-signer", required_signers),
            *helpers._prepend_flag("--required-signer-hash", required_signer_hashes),
            *helpers._prepend_flag("--certificate-file", tx_files.certificate_files),
            *helpers._prepend_flag(proposal_file_argname, tx_files.proposal_files),
            *helpers._prepend_flag("--vote-file", tx_files.vote_files),
            *helpers._prepend_flag("--auxiliary-script-file", tx_files.auxiliary_script_files),
            *helpers._prepend_flag("--metadata-json-file", tx_files.metadata_json_files),
            *helpers._prepend_flag("--metadata-cbor-file", tx_files.metadata_cbor_files),
            *helpers._prepend_flag("--withdrawal", withdrawal_strings),
            *txtools._get_return_collateral_txout_args(txouts=return_collateral_txouts),
            "--change-address",
            change_address,
            *misc_args,
            *self._clusterlib_obj.magic_args,
            *self._clusterlib_obj.socket_args,
        ]
        out = self._clusterlib_obj.cli(cli_args)
        stdout_dec = out.stdout.strip().decode("utf-8") if out.stdout else ""

        # Check for the presence of fee information. No fee information was provided in older
        # versions of the `build` command.
        estimated_fee = -1
        if stdout_dec.endswith("Lovelace"):
            estimated_fee = int(stdout_dec.split()[-2])
        elif "transaction fee" in stdout_dec:
            estimated_fee = int(stdout_dec.split()[-1])
        else:
            fee_str = self.view_tx_dict(tx_body_file=out_file).get("fee") or ""
            if fee_str.endswith("Lovelace"):
                estimated_fee = int(fee_str.split()[-2])

        return structs.TxRawOutput(
            txins=list(collected_data.txins),
            txouts=processed_txouts,
            txouts_count=txouts_count,
            tx_files=tx_files,
            out_file=out_file,
            fee=estimated_fee,
            build_args=cli_args,
            era=self._clusterlib_obj.era_in_use,
            script_txins=script_txins,
            script_withdrawals=collected_data.script_withdrawals,
            script_votes=script_votes,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            mint=mint,
            invalid_hereafter=invalid_hereafter,
            invalid_before=invalid_before,
            treasury_donation=treasury_donation,
            withdrawals=collected_data.withdrawals,
            change_address=change_address or src_address,
            return_collateral_txouts=return_collateral_txouts,
            total_collateral_amount=total_collateral_amount,
            readonly_reference_txins=readonly_reference_txins,
            script_valid=script_valid,
            required_signers=required_signers,
            required_signer_hashes=required_signer_hashes,
            combined_reference_txins=txtools._get_reference_txins(
                readonly_reference_txins=readonly_reference_txins,
                script_txins=script_txins,
                mint=mint,
                complex_certs=complex_certs,
                script_withdrawals=script_withdrawals,
            ),
        )

    def build_estimate_tx(  # noqa: C901
        self,
        src_address: str,
        tx_name: str,
        txins: structs.OptionalUTXOData = (),
        txouts: structs.OptionalTxOuts = (),
        readonly_reference_txins: structs.OptionalUTXOData = (),
        script_txins: structs.OptionalScriptTxIn = (),
        return_collateral_txouts: structs.OptionalTxOuts = (),
        total_collateral_amount: int | None = None,
        mint: structs.OptionalMint = (),
        tx_files: structs.TxFiles | None = None,
        complex_certs: structs.OptionalScriptCerts = (),
        complex_proposals: structs.OptionalScriptProposals = (),
        change_address: str = "",
        fee_buffer: int | None = None,
        required_signers: itp.OptionalFiles = (),
        required_signer_hashes: tp.Optional[list[str]] = None,
        withdrawals: structs.OptionalTxOuts = (),
        script_withdrawals: structs.OptionalScriptWithdrawals = (),
        script_votes: structs.OptionalScriptVotes = (),
        deposit: int | None = None,
        current_treasury_value: int | None = None,
        treasury_donation: int | None = None,
        invalid_hereafter: int | None = None,
        invalid_before: int | None = None,
        script_valid: bool = True,
        src_addr_utxos: list[structs.UTXOData] | None = None,
        witness_count_add: int = 0,
        byron_witness_count: int = 0,
        reference_script_size: int = 0,
        join_txouts: bool = True,
        destination_dir: itp.FileType = ".",
        skip_asset_balancing: bool = True,
    ) -> structs.TxRawOutput:
        """Build a balanced transaction without access to a live node.

        Args:
            src_address: An address used for fee and inputs (if inputs not specified by `txins`).
            tx_name: A name of the transaction.
            txins: An iterable of `structs.UTXOData`, specifying input UTxOs (optional).
            txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
            readonly_reference_txins: An iterable of `structs.UTXOData`, specifying input
                UTxOs to be referenced and used as readonly (optional).
            script_txins: An iterable of `ScriptTxIn`, specifying input script UTxOs (optional).
            return_collateral_txouts: A list (iterable) of `TxOuts`, specifying transaction outputs
                for excess collateral (optional).
            total_collateral_amount: An integer indicating the total amount of collateral
                (optional).
            mint: An iterable of `Mint`, specifying script minting data (optional).
            tx_files: A `structs.TxFiles` data container containing files needed for the transaction
                (optional).
            complex_certs: An iterable of `ComplexCert`, specifying certificates script data
                (optional).
            complex_proposals: An iterable of `ComplexProposal`, specifying proposals script data
                (optional).
            change_address: A string with address where ADA in excess of the transaction fee
                will go to (`src_address` by default).
            fee_buffer: A buffer for fee amount (optional).
            required_signers: An iterable of filepaths of the signing keys whose signatures
                are required (optional).
            required_signer_hashes: A list of hashes of the signing keys whose signatures
                are required (optional).
            withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals (optional).
            script_withdrawals: An iterable of `ScriptWithdrawal`, specifying withdrawal script
                data (optional).
            script_votes: An iterable of `ScriptVote`, specifying vote script data (optional).
            deposit: A deposit amount needed by the transaction (optional).
            current_treasury_value: The current treasury value (optional).
            treasury_donation: A donation to the treasury to perform (optional).
            invalid_hereafter: A last block when the transaction is still valid (optional).
            invalid_before: A first block when the transaction is valid (optional).
            script_valid: A bool indicating that the script is valid (True by default).
            src_addr_utxos: A list of UTxOs for the source address (optional).
            witness_count_add: A number of witnesses to add - workaround to make the fee
                calculation more precise.
            byron_witness_count: A number of Byron witnesses (optional).
            reference_script_size: A size in bytes of transaction reference scripts (optional).
            join_txouts: A bool indicating whether to aggregate transaction outputs
                by payment address (True by default).
            destination_dir: A path to directory for storing artifacts (optional).
            skip_asset_balancing: A bool indicating if assets balancing should be skipped.

        Returns:
            structs.TxRawOutput: A data container with transaction output details.
        """
        max_txout = [o for o in txouts if o.amount == -1 and o.coin in ("", consts.DEFAULT_COIN)]
        if max_txout:
            if change_address:
                msg = "Cannot use '-1' amount and change address at the same time."
                raise AssertionError(msg)
            change_address = max_txout[0].address
        else:
            change_address = change_address or src_address

        if (treasury_donation is not None) != (current_treasury_value is not None):
            msg = (
                "Both `treasury_donation` and `current_treasury_value` must be specified together."
            )
            raise AssertionError(msg)

        tx_files = tx_files or structs.TxFiles()
        if tx_files.certificate_files and complex_certs:
            LOGGER.warning(
                "Mixing `tx_files.certificate_files` and `complex_certs`, "
                "certs may come in unexpected order."
            )

        if tx_files.proposal_files and complex_proposals:
            LOGGER.warning(
                "Mixing `tx_files.proposal_files` and `complex_proposals`, "
                "proposals may come in unexpected order."
            )

        destination_dir = pl.Path(destination_dir).expanduser()

        out_file = destination_dir / f"{tx_name}_tx.body"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        collected_data = txtools.collect_data_for_build(
            clusterlib_obj=self._clusterlib_obj,
            src_address=src_address,
            txins=txins,
            txouts=txouts,
            script_txins=script_txins,
            mint=mint,
            tx_files=tx_files,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            fee=fee_buffer or 0,
            withdrawals=withdrawals,
            script_withdrawals=script_withdrawals,
            deposit=deposit,
            treasury_donation=treasury_donation,
            src_addr_utxos=src_addr_utxos,
            skip_asset_balancing=skip_asset_balancing,
        )

        required_signer_hashes = required_signer_hashes or []

        txout_args, processed_txouts, txouts_count = txtools._process_txouts(
            txouts=collected_data.txouts, join_txouts=join_txouts
        )

        txin_strings = txtools._get_txin_strings(
            txins=collected_data.txins, script_txins=script_txins
        )

        withdrawal_strings = [f"{x.address}+{x.amount}" for x in collected_data.withdrawals]

        mint_txouts = list(itertools.chain.from_iterable(m.txouts for m in mint))

        script_txins_records = list(itertools.chain.from_iterable(r.txins for r in script_txins))
        combined_txins = [
            *collected_data.txins,
            *script_txins_records,
        ]
        total_utxo_value = txtools.calculate_utxos_balance(utxos=combined_txins)

        estimate_args = [
            "--shelley-key-witnesses",
            str(len(tx_files.signing_key_files) + witness_count_add),
            "--byron-key-witnesses",
            str(byron_witness_count),
            "--reference-script-size",
            str(reference_script_size),
            "--total-utxo-value",
            str(total_utxo_value),
        ]

        misc_args = []

        if invalid_before is not None:
            misc_args.extend(["--invalid-before", str(invalid_before)])
        if invalid_hereafter is not None:
            misc_args.extend(["--invalid-hereafter", str(invalid_hereafter)])

        if treasury_donation is not None:
            misc_args.extend(["--treasury-donation", str(treasury_donation)])

        if not script_valid:
            misc_args.append("--script-invalid")

        # There's allowed just single `--mint` argument, let's aggregate all the outputs
        mint_records = [f"{m.amount} {m.coin}" for m in mint_txouts]
        misc_args.extend(["--mint", "+".join(mint_records)] if mint_records else [])

        for txin in readonly_reference_txins:
            misc_args.extend(["--read-only-tx-in-reference", f"{txin.utxo_hash}#{txin.utxo_ix}"])

        grouped_args = txtools._get_script_args(
            script_txins=script_txins,
            mint=mint,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            script_withdrawals=collected_data.script_withdrawals,
            script_votes=script_votes,
            with_execution_units=True,
        )

        if total_collateral_amount:
            misc_args.extend(["--tx-total-collateral", str(total_collateral_amount)])

        if tx_files.metadata_json_files and tx_files.metadata_json_detailed_schema:
            misc_args.append("--json-metadata-detailed-schema")

        self._clusterlib_obj.create_pparams_file()

        cli_args = [
            "transaction",
            "build-estimate",
            "--out-file",
            str(out_file),
            *estimate_args,
            *grouped_args,
            *helpers._prepend_flag("--tx-in", txin_strings),
            *txout_args,
            *helpers._prepend_flag("--required-signer", required_signers),
            *helpers._prepend_flag("--required-signer-hash", required_signer_hashes),
            *helpers._prepend_flag("--certificate-file", tx_files.certificate_files),
            *helpers._prepend_flag("--proposal-file", tx_files.proposal_files),
            *helpers._prepend_flag("--vote-file", tx_files.vote_files),
            *helpers._prepend_flag("--auxiliary-script-file", tx_files.auxiliary_script_files),
            *helpers._prepend_flag("--metadata-json-file", tx_files.metadata_json_files),
            *helpers._prepend_flag("--metadata-cbor-file", tx_files.metadata_cbor_files),
            *helpers._prepend_flag("--withdrawal", withdrawal_strings),
            *txtools._get_return_collateral_txout_args(txouts=return_collateral_txouts),
            "--change-address",
            change_address,
            "--protocol-params-file",
            str(self._clusterlib_obj.pparams_file),
            *misc_args,
            *self._clusterlib_obj.socket_args,
        ]
        out = self._clusterlib_obj.cli(cli_args)
        stdout_dec = out.stdout.strip().decode("utf-8") if out.stdout else ""

        # Check for the presence of fee information. No fee information was provided in older
        # versions of the `build-estimate` command. Try to get the fee information from the
        # `transaction view` command if not available from the `build-estimate`.
        estimated_fee = -1
        fee_str = stdout_dec
        if not fee_str.endswith("Lovelace"):
            fee_str = self.view_tx_dict(tx_body_file=out_file).get("fee") or ""
        if fee_str.endswith("Lovelace"):
            estimated_fee = int(fee_str.split()[-2])

        return structs.TxRawOutput(
            txins=list(collected_data.txins),
            txouts=processed_txouts,
            txouts_count=txouts_count,
            tx_files=tx_files,
            out_file=out_file,
            fee=estimated_fee,
            build_args=cli_args,
            era=self._clusterlib_obj.era_in_use,
            script_txins=script_txins,
            script_withdrawals=collected_data.script_withdrawals,
            script_votes=script_votes,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            mint=mint,
            invalid_hereafter=invalid_hereafter,
            invalid_before=invalid_before,
            treasury_donation=treasury_donation,
            withdrawals=collected_data.withdrawals,
            change_address=change_address or src_address,
            return_collateral_txouts=return_collateral_txouts,
            total_collateral_amount=total_collateral_amount,
            readonly_reference_txins=readonly_reference_txins,
            script_valid=script_valid,
            required_signers=required_signers,
            required_signer_hashes=required_signer_hashes,
            combined_reference_txins=txtools._get_reference_txins(
                readonly_reference_txins=readonly_reference_txins,
                script_txins=script_txins,
                mint=mint,
                complex_certs=complex_certs,
                script_withdrawals=script_withdrawals,
            ),
        )

    def sign_tx(
        self,
        signing_key_files: itp.OptionalFiles,
        tx_name: str,
        tx_body_file: itp.FileType | None = None,
        tx_file: itp.FileType | None = None,
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Sign a transaction.

        Args:
            signing_key_files: A list of paths to signing key files.
            tx_name: A name of the transaction.
            tx_body_file: A path to file with transaction body (optional).
            tx_file: A path to file with transaction (for incremental signing, optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to signed transaction file.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{tx_name}_tx.signed"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        if tx_body_file:
            cli_args = ["--tx-body-file", str(tx_body_file)]
        elif tx_file:
            cli_args = ["--tx-file", str(tx_file)]
        else:
            msg = "Either `tx_body_file` or `tx_file` is needed."
            raise AssertionError(msg)

        self._clusterlib_obj.cli(
            [
                "transaction",
                "sign",
                *cli_args,
                *self._clusterlib_obj.magic_args,
                *helpers._prepend_flag("--signing-key-file", signing_key_files),
                "--out-file",
                str(out_file),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def witness_tx(
        self,
        tx_body_file: itp.FileType,
        witness_name: str,
        signing_key_files: itp.OptionalFiles = (),
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Create a transaction witness.

        Args:
            tx_body_file: A path to file with transaction body.
            witness_name: A name of the transaction witness.
            signing_key_files: A list of paths to signing key files (optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to transaction witness file.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{witness_name}_tx.witness"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                "transaction",
                "witness",
                "--tx-body-file",
                str(tx_body_file),
                "--out-file",
                str(out_file),
                *self._clusterlib_obj.magic_args,
                *helpers._prepend_flag("--signing-key-file", signing_key_files),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def assemble_tx(
        self,
        tx_body_file: itp.FileType,
        witness_files: itp.OptionalFiles,
        tx_name: str,
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Assemble a tx body and witness(es) to form a signed transaction.

        Args:
            tx_body_file: A path to file with transaction body.
            witness_files: A list of paths to transaction witness files.
            tx_name: A name of the transaction.
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to signed transaction file.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{tx_name}_tx.witnessed"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        self._clusterlib_obj.cli(
            [
                "transaction",
                "assemble",
                "--tx-body-file",
                str(tx_body_file),
                "--out-file",
                str(out_file),
                *helpers._prepend_flag("--witness-file", witness_files),
            ]
        )

        helpers._check_outfiles(out_file)
        return out_file

    def submit_tx_bare(self, tx_file: itp.FileType) -> str:
        """Submit a transaction, don't do any verification that it made it to the chain.

        Args:
            tx_file: A path to signed transaction file.

        Returns:
            str: A transaction ID.
        """
        out = self._clusterlib_obj.cli(
            [
                "transaction",
                "submit",
                *self._clusterlib_obj.magic_args,
                *self._clusterlib_obj.socket_args,
                "--tx-file",
                str(tx_file),
            ]
        )

        stdout_dec = out.stdout.strip().decode("utf-8") if out.stdout else ""
        txhash_maybe = stdout_dec.split("\n")[-1]
        txhash = (json.loads(txhash_maybe).get("txhash") or "") if "txhash" in txhash_maybe else ""
        return txhash

    def submit_tx(
        self,
        tx_file: itp.FileType,
        txins: list[structs.UTXOData],
        wait_blocks: int | None = None,
    ) -> str:
        """Submit a transaction, resubmit if the transaction didn't make it to the chain.

        Args:
            tx_file: A path to signed transaction file.
            txins: An iterable of `structs.UTXOData`, specifying input UTxOs.
            wait_blocks: A number of new blocks to wait for (default = 2).

        Returns:
            str: A transaction ID.
        """
        wait_blocks = (
            self._clusterlib_obj.confirm_blocks
            if wait_blocks is None or wait_blocks < 1
            else wait_blocks
        )
        txid = ""
        for r in range(20):
            err = None

            if r == 0:
                txid = self.submit_tx_bare(tx_file)
            else:
                txid = txid or self.get_txid(tx_file=tx_file)
                LOGGER.warning(f"Resubmitting transaction '{txid}' (from '{tx_file}').")
                try:
                    self.submit_tx_bare(tx_file)
                except exceptions.CLIError as exc:
                    # Check if resubmitting failed because an input UTxO was already spent
                    if "(BadInputsUTxO" not in str(exc):
                        raise
                    err = err or exc
                    # If here, the TX is likely still in mempool and we need to wait

            self._clusterlib_obj.wait_for_new_block(wait_blocks)

            # Check that one of the input UTxOs can no longer be queried in order to verify
            # the TX was successfully submitted to the chain (that the TX is no longer in mempool).
            # An input is spent when its combination of hash and ix is not found in the list
            # of current UTxOs.
            # TODO: check that the transaction is 1-block deep (can't be done in CLI alone)
            utxo_data = self._clusterlib_obj.g_query.get_utxo(utxo=txins[0])
            if not utxo_data:
                break
        else:
            if err is not None:
                # Submitting the TX raised an exception as if the input was already
                # spent, but it was either not the case, or the TX is still in mempool.
                msg = f"Failed to resubmit the transaction '{txid}' (from '{tx_file}')."
                raise exceptions.CLIError(msg) from err

            msg = f"Transaction '{txid}' didn't make it to the chain (from '{tx_file}')."
            raise exceptions.CLIError(msg)

        return txid

    def send_tx(
        self,
        src_address: str,
        tx_name: str,
        txins: structs.OptionalUTXOData = (),
        txouts: structs.OptionalTxOuts = (),
        readonly_reference_txins: structs.OptionalUTXOData = (),
        script_txins: structs.OptionalScriptTxIn = (),
        return_collateral_txouts: structs.OptionalTxOuts = (),
        total_collateral_amount: int | None = None,
        mint: structs.OptionalMint = (),
        tx_files: structs.TxFiles | None = None,
        complex_certs: structs.OptionalScriptCerts = (),
        complex_proposals: structs.OptionalScriptProposals = (),
        fee: int | None = None,
        required_signers: itp.OptionalFiles = (),
        required_signer_hashes: tp.Optional[list[str]] = None,
        ttl: int | None = None,
        withdrawals: structs.OptionalTxOuts = (),
        script_withdrawals: structs.OptionalScriptWithdrawals = (),
        script_votes: structs.OptionalScriptVotes = (),
        deposit: int | None = None,
        current_treasury_value: int | None = None,
        treasury_donation: int | None = None,
        invalid_hereafter: int | None = None,
        invalid_before: int | None = None,
        witness_count_add: int = 0,
        join_txouts: bool = True,
        verify_tx: bool = True,
        destination_dir: itp.FileType = ".",
    ) -> structs.TxRawOutput:
        """Build, Sign and Submit a transaction.

        Not recommended for complex transactions that involve Plutus scripts!

        This function uses `cardano-cli transaction build-raw` to build the transaction.
        For more complex transactions that involve Plutus scripts, consider using `build_tx`.
        The `build_tx` uses `cardano-cli transaction build` and handles execution units and
        collateral return automatically.

        Args:
            src_address: An address used for fee and inputs (if inputs not specified by `txins`).
            tx_name: A name of the transaction.
            txins: An iterable of `structs.UTXOData`, specifying input UTxOs (optional).
            txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
            readonly_reference_txins: An iterable of `structs.UTXOData`, specifying input
                UTxOs to be referenced and used as readonly (optional).
            script_txins: An iterable of `ScriptTxIn`, specifying input script UTxOs (optional).
            return_collateral_txouts: A list (iterable) of `TxOuts`, specifying transaction outputs
                for excess collateral (optional).
            total_collateral_amount: An integer indicating the total amount of collateral
                (optional).
            mint: An iterable of `Mint`, specifying script minting data (optional).
            tx_files: A `structs.TxFiles` data container containing files needed for the transaction
                (optional).
            complex_certs: An iterable of `ComplexCert`, specifying certificates script data
                (optional).
            complex_proposals: An iterable of `ComplexProposal`, specifying proposal script data
                (optional).
            fee: A fee amount (optional).
            required_signers: An iterable of filepaths of the signing keys whose signatures
                are required (optional).
            required_signer_hashes: A list of hashes of the signing keys whose signatures
                are required (optional).
            ttl: A last block when the transaction is still valid
                (deprecated in favor of `invalid_hereafter`, optional).
            withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals (optional).
            script_withdrawals: An iterable of `ScriptWithdrawal`, specifying withdrawal script
                data (optional).
            script_votes: An iterable of `ScriptVote`, specifying vote script data (optional).
            deposit: A deposit amount needed by the transaction (optional).
            current_treasury_value: The current treasury value (optional).
            treasury_donation: A donation to the treasury to perform (optional).
            invalid_hereafter: A last block when the transaction is still valid (optional).
            invalid_before: A first block when the transaction is valid (optional).
            witness_count_add: A number of witnesses to add - workaround to make the fee
                calculation more precise.
            join_txouts: A bool indicating whether to aggregate transaction outputs
                by payment address (True by default).
            verify_tx: A bool indicating whether to verify the transaction made it to chain
                and resubmit the transaction if not (True by default).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            structs.TxRawOutput: A data container with transaction output details.
        """
        tx_files = tx_files or structs.TxFiles()

        # Resolve withdrawal amounts here (where -1 for total rewards amount is used) so the
        # resolved values can be passed around, and it is not needed to resolve them again
        # every time `_get_withdrawals` is called
        withdrawals, script_withdrawals, *__ = txtools._get_withdrawals(
            clusterlib_obj=self._clusterlib_obj,
            withdrawals=withdrawals,
            script_withdrawals=script_withdrawals,
        )

        # Get UTxOs for src address here so the records can be passed around, and it is
        # not necessary to get them once for fee calculation and again for the final transaction
        # building.
        src_addr_utxos = (
            self._clusterlib_obj.g_query.get_utxo(address=src_address)
            if fee is None and not txins
            else None
        )

        if fee is None:
            fee = self.calculate_tx_fee(
                src_address=src_address,
                tx_name=tx_name,
                txins=txins,
                txouts=txouts,
                readonly_reference_txins=readonly_reference_txins,
                script_txins=script_txins,
                return_collateral_txouts=return_collateral_txouts,
                total_collateral_amount=total_collateral_amount,
                mint=mint,
                tx_files=tx_files,
                complex_certs=complex_certs,
                complex_proposals=complex_proposals,
                required_signers=required_signers,
                required_signer_hashes=required_signer_hashes,
                withdrawals=withdrawals,
                script_withdrawals=script_withdrawals,
                deposit=deposit,
                current_treasury_value=current_treasury_value,
                treasury_donation=treasury_donation,
                invalid_hereafter=invalid_hereafter or ttl,
                src_addr_utxos=src_addr_utxos,
                witness_count_add=witness_count_add,
                join_txouts=join_txouts,
                destination_dir=destination_dir,
            )

            # Add 10% to the estimated fee, as the estimation is not precise enough, and there
            # might be another txin in the final tx once fee is added to the total needed amount
            fee = int(fee * 1.1)

        tx_raw_output = self.build_raw_tx(
            src_address=src_address,
            tx_name=tx_name,
            txins=txins,
            txouts=txouts,
            readonly_reference_txins=readonly_reference_txins,
            script_txins=script_txins,
            return_collateral_txouts=return_collateral_txouts,
            total_collateral_amount=total_collateral_amount,
            mint=mint,
            tx_files=tx_files,
            complex_certs=complex_certs,
            complex_proposals=complex_proposals,
            fee=fee,
            required_signers=required_signers,
            required_signer_hashes=required_signer_hashes,
            withdrawals=withdrawals,
            script_withdrawals=script_withdrawals,
            script_votes=script_votes,
            deposit=deposit,
            current_treasury_value=current_treasury_value,
            treasury_donation=treasury_donation,
            invalid_hereafter=invalid_hereafter or ttl,
            invalid_before=invalid_before,
            src_addr_utxos=src_addr_utxos,
            join_txouts=join_txouts,
            destination_dir=destination_dir,
        )
        tx_signed_file = self.sign_tx(
            tx_body_file=tx_raw_output.out_file,
            tx_name=tx_name,
            signing_key_files=tx_files.signing_key_files,
            destination_dir=destination_dir,
        )
        if verify_tx:
            self.submit_tx(
                tx_file=tx_signed_file,
                txins=tx_raw_output.txins
                or [t.txins[0] for t in tx_raw_output.script_txins if t.txins],
            )
        else:
            self.submit_tx_bare(tx_file=tx_signed_file)

        return tx_raw_output

    def build_multisig_script(
        self,
        script_name: str,
        script_type_arg: str,
        payment_vkey_files: itp.OptionalFiles,
        required: int = 0,
        slot: int = 0,
        slot_type_arg: str = "",
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Build a multi-signature script.

        Args:
            script_name: A name of the script.
            script_type_arg: A script type, see `MultiSigTypeArgs`.
            payment_vkey_files: A list of paths to payment vkey files.
            required: A number of required keys for the "atLeast" script type (optional).
            slot: A slot that sets script validity, depending on value of `slot_type_arg`
                (optional).
            slot_type_arg: A slot validity type, see `MultiSlotTypeArgs` (optional).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            Path: A path to the script file.
        """
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{script_name}_multisig.script"

        scripts_l: list[dict] = [
            {
                "keyHash": self._clusterlib_obj.g_address.get_payment_vkey_hash(
                    payment_vkey_file=f
                ),
                "type": "sig",
            }
            for f in payment_vkey_files
        ]
        if slot:
            scripts_l.append({"slot": slot, "type": slot_type_arg})

        script: dict = {
            "scripts": scripts_l,
            "type": script_type_arg,
        }

        if script_type_arg == consts.MultiSigTypeArgs.AT_LEAST:
            script["required"] = required

        with open(out_file, "w", encoding="utf-8") as fp_out:
            json.dump(script, fp_out, indent=4)

        return out_file

    def get_policyid(
        self,
        script_file: itp.FileType,
    ) -> str:
        """Calculate the PolicyId from the monetary policy script.

        Args:
            script_file: A path to the script file.

        Returns:
            str: A script policyId.
        """
        return (
            self._clusterlib_obj.cli(["transaction", "policyid", "--script-file", str(script_file)])
            .stdout.rstrip()
            .decode("utf-8")
        )

    def calculate_plutus_script_cost(
        self,
        src_address: str,
        tx_name: str,
        txins: structs.OptionalUTXOData = (),
        txouts: structs.OptionalTxOuts = (),
        readonly_reference_txins: structs.OptionalUTXOData = (),
        script_txins: structs.OptionalScriptTxIn = (),
        return_collateral_txouts: structs.OptionalTxOuts = (),
        total_collateral_amount: int | None = None,
        mint: structs.OptionalMint = (),
        tx_files: structs.TxFiles | None = None,
        complex_certs: structs.OptionalScriptCerts = (),
        complex_proposals: structs.OptionalScriptProposals = (),
        change_address: str = "",
        fee_buffer: int | None = None,
        required_signers: itp.OptionalFiles = (),
        required_signer_hashes: tp.Optional[list[str]] = None,
        withdrawals: structs.OptionalTxOuts = (),
        script_withdrawals: structs.OptionalScriptWithdrawals = (),
        script_votes: structs.OptionalScriptVotes = (),
        deposit: int | None = None,
        invalid_hereafter: int | None = None,
        invalid_before: int | None = None,
        witness_override: int | None = None,
        script_valid: bool = True,
        calc_script_cost_file: itp.FileType | None = None,
        join_txouts: bool = True,
        destination_dir: itp.FileType = ".",
    ) -> list[dict]:
        """Calculate cost of Plutus scripts. Accepts the same arguments as `build_tx`.

        Args:
            src_address: An address used for fee and inputs (if inputs not specified by `txins`).
            tx_name: A name of the transaction.
            txins: An iterable of `structs.UTXOData`, specifying input UTxOs (optional).
            txouts: A list (iterable) of `TxOuts`, specifying transaction outputs (optional).
            readonly_reference_txins: An iterable of `structs.UTXOData`, specifying input
                UTxOs to be referenced and used as readonly (optional).
            script_txins: An iterable of `ScriptTxIn`, specifying input script UTxOs (optional).
            return_collateral_txouts: A list (iterable) of `TxOuts`, specifying transaction outputs
                for excess collateral (optional).
            total_collateral_amount: An integer indicating the total amount of collateral
                (optional).
            mint: An iterable of `Mint`, specifying script minting data (optional).
            tx_files: A `structs.TxFiles` data container containing files needed for the transaction
                (optional).
            complex_certs: An iterable of `ComplexCert`, specifying certificates script data
                (optional).
            complex_proposals: An iterable of `ComplexProposal`, specifying proposal script data
                (optional).
            change_address: A string with address where ADA in excess of the transaction fee
                will go to (`src_address` by default).
            fee_buffer: A buffer for fee amount (optional).
            required_signers: An iterable of filepaths of the signing keys whose signatures
                are required (optional).
            required_signer_hashes: A list of hashes of the signing keys whose signatures
                are required (optional).
            withdrawals: A list (iterable) of `TxOuts`, specifying reward withdrawals (optional).
            script_withdrawals: An iterable of `ScriptWithdrawal`, specifying withdrawal script
                data (optional).
            script_votes: An iterable of `ScriptVote`, specifying vote script data (optional).
            deposit: A deposit amount needed by the transaction (optional).
            invalid_hereafter: A last block when the transaction is still valid (optional).
            invalid_before: A first block when the transaction is valid (optional).
            witness_override: An integer indicating real number of witnesses. Can be used to fix
                fee calculation (optional).
            script_valid: A bool indicating that the script is valid (True by default).
            calc_script_cost_file: A path for output of the Plutus script cost information
                (optional).
            join_txouts: A bool indicating whether to aggregate transaction outputs
                by payment address (True by default).
            destination_dir: A path to directory for storing artifacts (optional).

        Returns:
            List[dict]: A Plutus scripts cost data.
        """
        # Collect all arguments that will be passed to `build_tx`
        kwargs = locals()
        kwargs.pop("self", None)
        kwargs.pop("kwargs", None)
        # This would be a duplicate if already present
        kwargs.pop("calc_script_cost_file", None)

        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{tx_name}_plutus.cost"

        self.build_tx(**kwargs, calc_script_cost_file=out_file)
        with open(out_file, encoding="utf-8") as fp_out:
            cost: list[dict] = json.load(fp_out)
        return cost

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: clusterlib_obj={id(self._clusterlib_obj)}>"



"""Helper functions for `ClusterLib`."""

import dataclasses
import datetime
import json
import logging
import pathlib as pl
import re
import time
import typing as tp

from cardano_clusterlib import exceptions
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)

SPECIAL_ARG_CHARS_RE = re.compile("[^A-Za-z0-9/._-]")


@dataclasses.dataclass(frozen=True, order=True)
class EpochInfo:
    epoch: int
    first_slot: int
    last_slot: int


def _find_genesis_json(clusterlib_obj: "itp.ClusterLib") -> pl.Path:
    """Find Shelley genesis JSON file in state dir."""
    default = clusterlib_obj.state_dir / "shelley" / "genesis.json"
    if default.exists():
        return default

    potential = [
        *clusterlib_obj.state_dir.glob("*shelley*genesis.json"),
        *clusterlib_obj.state_dir.glob("*genesis*shelley.json"),
    ]
    if not potential:
        msg = f"Shelley genesis JSON file not found in `{clusterlib_obj.state_dir}`."
        raise exceptions.CLIError(msg)

    genesis_json = potential[0]
    LOGGER.debug(f"Using shelley genesis JSON file `{genesis_json}")
    return genesis_json


def _find_conway_genesis_json(clusterlib_obj: "itp.ClusterLib") -> pl.Path:
    """Find Conway genesis JSON file in state dir."""
    default = clusterlib_obj.state_dir / "shelley" / "genesis.conway.json"
    if default.exists():
        return default

    potential = [
        *clusterlib_obj.state_dir.glob("*conway*genesis.json"),
        *clusterlib_obj.state_dir.glob("*genesis*conway.json"),
    ]
    if not potential:
        msg = f"Conway genesis JSON file not found in `{clusterlib_obj.state_dir}`."
        raise exceptions.CLIError(msg)

    genesis_json = potential[0]
    LOGGER.debug(f"Using Conway genesis JSON file `{genesis_json}")
    return genesis_json


def _check_files_exist(*out_files: itp.FileType, clusterlib_obj: "itp.ClusterLib") -> None:
    """Check that the output files don't already exist.

    Args:
        *out_files: Variable length list of expected output files.
        clusterlib_obj: An instance of `ClusterLib`.
    """
    if clusterlib_obj.overwrite_outfiles:
        return

    for out_file in out_files:
        out_file_p = pl.Path(out_file).expanduser()
        if out_file_p.exists():
            msg = f"The expected file `{out_file}` already exist."
            raise exceptions.CLIError(msg)


def _format_cli_args(cli_args: list[str]) -> str:
    """Format CLI arguments for logging.

    Quote arguments with spaces and other "special" characters in them.

    Args:
        cli_args: List of CLI arguments.
    """
    processed_args = []
    for arg in cli_args:
        arg_p = f'"{arg}"' if SPECIAL_ARG_CHARS_RE.search(arg) else arg
        processed_args.append(arg_p)
    return " ".join(processed_args)


def _write_cli_log(clusterlib_obj: "itp.ClusterLib", command: str) -> None:
    if not clusterlib_obj._cli_log:
        return

    with open(clusterlib_obj._cli_log, "a", encoding="utf-8") as logfile:
        logfile.write(f"{datetime.datetime.now(tz=datetime.timezone.utc)}: {command}\n")


def _get_kes_period_info(kes_info: str) -> dict[str, tp.Any]:
    """Process the output of the `kes-period-info` command.

    Args:
        kes_info: The output of the `kes-period-info` command.
    """
    messages_str = kes_info.split("{")[0]
    messages_list = []

    valid_counters = False
    valid_kes_period = False

    if messages_str:
        message_entry: list = []

        for line in messages_str.split("\n"):
            line_s = line.strip()
            if not line_s:
                continue
            if not message_entry or line_s[0].isalpha():
                message_entry.append(line_s)
            else:
                messages_list.append(" ".join(message_entry))
                message_entry = [line_s]

        messages_list.append(" ".join(message_entry))

        for out_message in messages_list:
            if (
                "counter agrees with" in out_message
                or "counter ahead of the node protocol state counter by 1" in out_message
            ):
                valid_counters = True
            elif "correct KES period interval" in out_message:
                valid_kes_period = True

    # Get output metrics
    metrics_str = kes_info.split("{")[-1]
    metrics_dict = {}

    if metrics_str and metrics_str.strip().endswith("}"):
        metrics_dict = json.loads(f"{{{metrics_str}")

    output_dict = {
        "messages": messages_list,
        "metrics": metrics_dict,
        "valid_counters": valid_counters,
        "valid_kes_period": valid_kes_period,
    }

    return output_dict


def get_epoch_for_slot(cluster_obj: "itp.ClusterLib", slot_no: int) -> EpochInfo:
    """Given slot number, return corresponding epoch number and first and last slot of the epoch."""
    genesis_byron = cluster_obj.state_dir / "byron" / "genesis.json"
    if not genesis_byron.exists():
        msg = f"File '{genesis_byron}' does not exist."
        raise AssertionError(msg)

    with open(genesis_byron, encoding="utf-8") as in_json:
        byron_dict = json.load(in_json)

    byron_k = int(byron_dict["protocolConsts"]["k"])
    slots_in_byron_epoch = byron_k * 10
    slots_per_epoch_diff = cluster_obj.epoch_length - slots_in_byron_epoch
    num_byron_epochs = cluster_obj.slots_offset // slots_per_epoch_diff
    slots_in_byron = num_byron_epochs * slots_in_byron_epoch

    # Slot is in Byron era
    if slot_no < slots_in_byron:
        epoch_no = slot_no // slots_in_byron_epoch
        first_slot_in_epoch = epoch_no * slots_in_byron_epoch
        last_slot_in_epoch = first_slot_in_epoch + slots_in_byron_epoch - 1
    # Slot is in Shelley-based era
    else:
        slot_no_shelley = slot_no + cluster_obj.slots_offset
        epoch_no = slot_no_shelley // cluster_obj.epoch_length
        first_slot_in_epoch = epoch_no * cluster_obj.epoch_length - cluster_obj.slots_offset
        last_slot_in_epoch = first_slot_in_epoch + cluster_obj.epoch_length - 1

    return EpochInfo(epoch=epoch_no, first_slot=first_slot_in_epoch, last_slot=last_slot_in_epoch)


def wait_for_block(clusterlib_obj: "itp.ClusterLib", tip: dict[str, tp.Any], block_no: int) -> int:
    """Wait for block number.

    Args:
        clusterlib_obj: An instance of `ClusterLib`.
        tip: Current tip - last block successfully applied to the ledger.
        block_no: A block number to wait for.

    Returns:
        int: A block number of last added block.
    """
    initial_block = int(tip["block"])
    initial_slot = int(tip["slot"])

    if initial_block >= block_no:
        return initial_block

    next_block_timeout = 300  # in slots
    max_tip_throttle = 5 * clusterlib_obj.slot_length

    new_blocks = block_no - initial_block

    LOGGER.debug(f"Waiting for {new_blocks} new block(s) to be created.")
    LOGGER.debug(f"Initial block no: {initial_block}")

    this_slot = initial_slot
    this_block = initial_block
    timeout_slot = initial_slot + next_block_timeout
    blocks_to_go = new_blocks
    # Limit calls to `query tip`
    tip_throttle = 0

    while this_slot < timeout_slot:
        prev_block = this_block
        time.sleep((clusterlib_obj.slot_length * blocks_to_go) + tip_throttle)

        this_tip = clusterlib_obj.g_query.get_tip()
        this_slot = int(this_tip["slot"])
        this_block = int(this_tip["block"])

        if this_block >= block_no:
            break
        if this_block > prev_block:
            # New block was created, reset timeout slot
            timeout_slot = this_slot + next_block_timeout

        blocks_to_go = block_no - this_block
        tip_throttle = min(max_tip_throttle, tip_throttle + clusterlib_obj.slot_length)
    else:
        waited_sec = (this_slot - initial_slot) * clusterlib_obj.slot_length
        msg = f"Timeout waiting for {waited_sec} sec for {new_blocks} block(s)."
        raise exceptions.CLIError(msg)

    LOGGER.debug(f"New block(s) were created; block number: {this_block}")
    return this_block


def poll_new_epoch(
    clusterlib_obj: "itp.ClusterLib",
    exp_epoch: int,
    padding_seconds: int = 0,
) -> None:
    """Wait for new epoch(s) by polling current epoch every 3 sec.

    Can be used only for waiting up to 3000 sec + padding seconds.

    Args:
        clusterlib_obj: An instance of `ClusterLib`.
        tip: Current tip - last block successfully applied to the ledger.
        exp_epoch: An epoch number to wait for.
        padding_seconds: A number of additional seconds to wait for (optional).
    """
    for check_no in range(1000):
        wakeup_epoch = clusterlib_obj.g_query.get_epoch()
        if wakeup_epoch != exp_epoch:
            time.sleep(3)
            continue
        # We are in the expected epoch right from the beginning, we'll skip padding seconds
        if check_no == 0:
            break
        if padding_seconds:
            time.sleep(padding_seconds)
            break


def wait_for_epoch(
    clusterlib_obj: "itp.ClusterLib",
    tip: dict[str, tp.Any],
    epoch_no: int,
    padding_seconds: int = 0,
    future_is_ok: bool = True,
) -> int:
    """Wait for epoch no.

    Args:
        clusterlib_obj: An instance of `ClusterLib`.
        tip: Current tip - last block successfully applied to the ledger.
        epoch_no: A number of epoch to wait for.
        padding_seconds: A number of additional seconds to wait for (optional).
        future_is_ok: A bool indicating whether current epoch > `epoch_no` is acceptable
            (default: True).

    Returns:
        int: The current epoch.
    """
    start_epoch = int(tip["epoch"])

    if epoch_no < start_epoch:
        if not future_is_ok:
            msg = f"Current epoch is {start_epoch}. The requested epoch {epoch_no} is in the past."
            raise exceptions.CLIError(msg)
        return start_epoch

    LOGGER.debug(f"Current epoch: {start_epoch}; Waiting for the beginning of epoch: {epoch_no}")

    new_epochs = epoch_no - start_epoch

    # Calculate and wait for the expected slot
    boundary_slot = int(
        (start_epoch + new_epochs) * clusterlib_obj.epoch_length - clusterlib_obj.slots_offset
    )
    padding_slots = int(padding_seconds / clusterlib_obj.slot_length) if padding_seconds else 5
    exp_slot = boundary_slot + padding_slots
    clusterlib_obj.wait_for_slot(slot=exp_slot)

    this_epoch = clusterlib_obj.g_query.get_epoch()
    if this_epoch != epoch_no:
        LOGGER.error(
            f"Waited for epoch number {epoch_no} and current epoch is "
            f"number {this_epoch}, wrong `slots_offset` ({clusterlib_obj.slots_offset})?"
        )
        # Attempt to get the epoch boundary as precisely as possible failed, now just
        # query epoch number and wait
        poll_new_epoch(
            clusterlib_obj=clusterlib_obj, exp_epoch=epoch_no, padding_seconds=padding_seconds
        )

    # Still not in the correct epoch? Something is wrong.
    this_epoch = clusterlib_obj.g_query.get_epoch()
    if this_epoch != epoch_no:
        msg = f"Waited for epoch number {epoch_no} and current epoch is number {this_epoch}."
        raise exceptions.CLIError(msg)

    LOGGER.debug(f"Expected epoch started; epoch number: {this_epoch}")
    return this_epoch


def get_slots_offset(clusterlib_obj: "itp.ClusterLib") -> int:
    """Get offset of slots from Byron era vs current configuration."""
    tip = clusterlib_obj.g_query.get_tip()
    slot = int(tip["slot"])
    slots_ep_end = int(tip["slotsToEpochEnd"])
    epoch = int(tip["epoch"])

    slots_total = slot + slots_ep_end
    slots_shelley = int(clusterlib_obj.epoch_length) * (epoch + 1)

    offset = slots_shelley - slots_total
    return offset



"""Group of methods for Conway governance action commands."""

import json
import logging
import pathlib as pl
import typing as tp

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import consts
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class GovActionGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj
        self._group_args = ("governance", "action")

        if self._clusterlib_obj.network_magic == consts.MAINNET_MAGIC:
            self.magic_args = ["--mainnet"]
        else:
            self.magic_args = ["--testnet"]

    def _get_deposit_return_key_args(
        self,
        deposit_return_stake_vkey: str = "",
        deposit_return_stake_vkey_file: itp.FileType | None = None,
        deposit_return_stake_key_hash: str = "",
    ) -> list[str]:
        """Get arguments for verification key."""
        if deposit_return_stake_vkey:
            key_args = ["--deposit-return-stake-verification-key", str(deposit_return_stake_vkey)]
        elif deposit_return_stake_vkey_file:
            key_args = [
                "--deposit-return-stake-verification-key-file",
                str(deposit_return_stake_vkey_file),
            ]
        elif deposit_return_stake_key_hash:
            key_args = ["--deposit-return-stake-key-hash", str(deposit_return_stake_key_hash)]
        else:
            msg = "Either stake verification key or stake key hash must be set."
            raise AssertionError(msg)

        return key_args

    def _get_cc_members_args(
        self,
        cc_members: list[structs.CCMember],
        remove: bool = False,
    ) -> list[str]:
        """Get arguments for committee members."""
        arg_action = "remove" if remove else "add"
        cc_members_args = []

        for cc_member in cc_members:
            if cc_member.cold_vkey:
                cc_members_args.extend(
                    [
                        f"--{arg_action}-cc-cold-verification-key",
                        str(cc_member.cold_vkey),
                    ]
                )
            elif cc_member.cold_vkey_file:
                cc_members_args.extend(
                    [
                        f"--{arg_action}-cc-cold-verification-key-file",
                        str(cc_member.cold_vkey_file),
                    ]
                )
            elif cc_member.cold_vkey_hash:
                cc_members_args.extend(
                    [
                        f"--{arg_action}-cc-cold-verification-key-hash",
                        str(cc_member.cold_vkey_hash),
                    ]
                )
            elif cc_member.cold_script_hash:
                cc_members_args.extend(
                    [
                        f"--{arg_action}-cc-cold-script-hash",
                        str(cc_member.cold_script_hash),
                    ]
                )
            else:
                msg = f"Either {arg_action} cold verification key or its hash must be set."
                raise AssertionError(msg)

            if not remove:
                cc_members_args.extend(["--epoch", str(cc_member.epoch)])

        return cc_members_args

    def _get_optional_prev_action_args(
        self,
        prev_action_txid: str = "",
        prev_action_ix: int = -1,
    ) -> list[str]:
        """Get arguments for previous action."""
        prev_action_args = []
        if prev_action_txid:
            if prev_action_ix == -1:
                msg = "Previous action index must be set."
                raise AssertionError(msg)
            prev_action_args = [
                "--prev-governance-action-tx-id",
                str(prev_action_txid),
                "--prev-governance-action-index",
                str(prev_action_ix),
            ]

        return prev_action_args

    def _get_anchor_args(
        self,
        anchor_url: str,
        anchor_data_hash: str = "",
    ) -> list[str]:
        """Get arguments for anchor."""
        anchor_args = [
            "--anchor-url",
            str(anchor_url),
            "--anchor-data-hash",
            str(anchor_data_hash),
        ]

        return anchor_args

    def create_constitution(
        self,
        action_name: str,
        deposit_amt: int,
        anchor_url: str,
        anchor_data_hash: str,
        constitution_url: str,
        constitution_hash: str,
        constitution_script_hash: str = "",
        deposit_return_stake_vkey: str = "",
        deposit_return_stake_vkey_file: itp.FileType | None = None,
        deposit_return_stake_key_hash: str = "",
        prev_action_txid: str = "",
        prev_action_ix: int = -1,
        destination_dir: itp.FileType = ".",
    ) -> structs.ActionConstitution:
        """Create a constitution."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{action_name}_constitution.action"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        key_args = self._get_deposit_return_key_args(
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=deposit_return_stake_vkey_file,
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )

        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )

        prev_action_args = self._get_optional_prev_action_args(
            prev_action_txid=prev_action_txid, prev_action_ix=prev_action_ix
        )

        constitution_anchor_args = [
            "--constitution-url",
            str(constitution_url),
            "--constitution-hash",
            str(constitution_hash),
        ]
        if constitution_script_hash:
            constitution_anchor_args.extend(
                [
                    "--constitution-script-hash",
                    str(constitution_script_hash),
                ]
            )

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create-constitution",
                *self.magic_args,
                "--governance-action-deposit",
                str(deposit_amt),
                *key_args,
                *prev_action_args,
                *anchor_args,
                *constitution_anchor_args,
                "--out-file",
                str(out_file),
            ]
        )
        helpers._check_outfiles(out_file)

        action_out = structs.ActionConstitution(
            action_file=out_file,
            deposit_amt=deposit_amt,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
            constitution_url=constitution_url,
            constitution_hash=constitution_hash,
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=helpers._maybe_path(file=deposit_return_stake_vkey_file),
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
            prev_action_txid=prev_action_txid,
            prev_action_ix=prev_action_ix,
        )

        return action_out

    def create_info(
        self,
        action_name: str,
        deposit_amt: int,
        anchor_url: str,
        anchor_data_hash: str,
        deposit_return_stake_vkey: str = "",
        deposit_return_stake_vkey_file: itp.FileType | None = None,
        deposit_return_stake_key_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> structs.ActionInfo:
        """Create an info action."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{action_name}_info.action"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        key_args = self._get_deposit_return_key_args(
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=deposit_return_stake_vkey_file,
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )

        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create-info",
                *self.magic_args,
                "--governance-action-deposit",
                str(deposit_amt),
                *key_args,
                *anchor_args,
                "--out-file",
                str(out_file),
            ]
        )
        helpers._check_outfiles(out_file)

        action_out = structs.ActionInfo(
            action_file=out_file,
            deposit_amt=deposit_amt,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=helpers._maybe_path(deposit_return_stake_vkey_file),
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )
        return action_out

    def create_no_confidence(
        self,
        action_name: str,
        deposit_amt: int,
        anchor_url: str,
        anchor_data_hash: str,
        prev_action_txid: str,
        prev_action_ix: int,
        deposit_return_stake_vkey: str = "",
        deposit_return_stake_vkey_file: itp.FileType | None = None,
        deposit_return_stake_key_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> structs.ActionNoConfidence:
        """Create a no confidence proposal."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{action_name}_confidence.action"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        key_args = self._get_deposit_return_key_args(
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=deposit_return_stake_vkey_file,
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )

        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )

        prev_action_args = [
            "--prev-governance-action-tx-id",
            str(prev_action_txid),
            "--prev-governance-action-index",
            str(prev_action_ix),
        ]

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create-no-confidence",
                *self.magic_args,
                "--governance-action-deposit",
                str(deposit_amt),
                *key_args,
                *prev_action_args,
                *anchor_args,
                "--out-file",
                str(out_file),
            ]
        )
        helpers._check_outfiles(out_file)

        action_out = structs.ActionNoConfidence(
            action_file=out_file,
            deposit_amt=deposit_amt,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
            prev_action_txid=prev_action_txid,
            prev_action_ix=prev_action_ix,
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=helpers._maybe_path(deposit_return_stake_vkey_file),
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )
        return action_out

    def update_committee(
        self,
        action_name: str,
        deposit_amt: int,
        anchor_url: str,
        anchor_data_hash: str,
        threshold: str,
        add_cc_members: tp.Optional[list[structs.CCMember]] = None,
        rem_cc_members: tp.Optional[list[structs.CCMember]] = None,
        prev_action_txid: str = "",
        prev_action_ix: int = -1,
        deposit_return_stake_vkey: str = "",
        deposit_return_stake_vkey_file: itp.FileType | None = None,
        deposit_return_stake_key_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> structs.ActionUpdateCommittee:
        """Create or update a new committee proposal."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{action_name}_update_committee.action"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )

        key_args = self._get_deposit_return_key_args(
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=deposit_return_stake_vkey_file,
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )

        prev_action_args = self._get_optional_prev_action_args(
            prev_action_txid=prev_action_txid, prev_action_ix=prev_action_ix
        )

        rem_cc_members_args = self._get_cc_members_args(
            cc_members=rem_cc_members or [], remove=True
        )
        add_cc_members_args = self._get_cc_members_args(cc_members=add_cc_members or [])

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "update-committee",
                *self.magic_args,
                "--governance-action-deposit",
                str(deposit_amt),
                *key_args,
                *prev_action_args,
                *anchor_args,
                *rem_cc_members_args,
                *add_cc_members_args,
                "--threshold",
                threshold,
                "--out-file",
                str(out_file),
            ]
        )
        helpers._check_outfiles(out_file)

        action_out = structs.ActionUpdateCommittee(
            action_file=out_file,
            deposit_amt=deposit_amt,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
            threshold=threshold,
            add_cc_members=add_cc_members or [],
            rem_cc_members=rem_cc_members or [],
            prev_action_txid=prev_action_txid,
            prev_action_ix=prev_action_ix,
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=helpers._maybe_path(deposit_return_stake_vkey_file),
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )
        return action_out

    def create_pparams_update(
        self,
        action_name: str,
        deposit_amt: int,
        anchor_url: str,
        anchor_data_hash: str,
        cli_args: itp.UnpackableSequence,
        prev_action_txid: str = "",
        prev_action_ix: int = -1,
        deposit_return_stake_vkey: str = "",
        deposit_return_stake_vkey_file: itp.FileType | None = None,
        deposit_return_stake_key_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> structs.ActionPParamsUpdate:
        """Create a protocol parameters update proposal."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{action_name}_pparams_update.action"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        key_args = self._get_deposit_return_key_args(
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=deposit_return_stake_vkey_file,
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )

        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )

        prev_action_args = self._get_optional_prev_action_args(
            prev_action_txid=prev_action_txid, prev_action_ix=prev_action_ix
        )

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create-protocol-parameters-update",
                *self.magic_args,
                "--governance-action-deposit",
                str(deposit_amt),
                *key_args,
                *prev_action_args,
                *anchor_args,
                *cli_args,
                "--out-file",
                str(out_file),
            ]
        )
        helpers._check_outfiles(out_file)

        action_out = structs.ActionPParamsUpdate(
            action_file=out_file,
            deposit_amt=deposit_amt,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
            cli_args=cli_args,
            prev_action_txid=prev_action_txid,
            prev_action_ix=prev_action_ix,
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=helpers._maybe_path(deposit_return_stake_vkey_file),
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )
        return action_out

    def create_treasury_withdrawal(
        self,
        action_name: str,
        transfer_amt: int,
        deposit_amt: int,
        anchor_url: str,
        anchor_data_hash: str,
        funds_receiving_stake_vkey: str = "",
        funds_receiving_stake_vkey_file: itp.FileType | None = None,
        funds_receiving_stake_key_hash: str = "",
        deposit_return_stake_vkey: str = "",
        deposit_return_stake_vkey_file: itp.FileType | None = None,
        deposit_return_stake_key_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> structs.ActionTreasuryWithdrawal:
        """Create a treasury withdrawal."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{action_name}_info.action"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        if funds_receiving_stake_vkey:
            funds_key_args = [
                "--funds-receiving-stake-verification-key",
                str(funds_receiving_stake_vkey),
            ]
        elif funds_receiving_stake_vkey_file:
            funds_key_args = [
                "--funds-receiving-stake-verification-key-file",
                str(funds_receiving_stake_vkey_file),
            ]
        elif funds_receiving_stake_key_hash:
            funds_key_args = [
                "--funds-receiving-stake-key-hash",
                str(funds_receiving_stake_key_hash),
            ]
        else:
            msg = "Either stake verification key or stake key hash must be set."
            raise AssertionError(msg)

        deposit_key_args = self._get_deposit_return_key_args(
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=deposit_return_stake_vkey_file,
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )

        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )

        self._clusterlib_obj.cli(
            [
                *self._group_args,
                "create-treasury-withdrawal",
                *self.magic_args,
                "--governance-action-deposit",
                str(deposit_amt),
                *deposit_key_args,
                *anchor_args,
                *funds_key_args,
                "--transfer",
                str(transfer_amt),
                "--out-file",
                str(out_file),
            ]
        )
        helpers._check_outfiles(out_file)

        action_out = structs.ActionTreasuryWithdrawal(
            action_file=out_file,
            transfer_amt=transfer_amt,
            deposit_amt=deposit_amt,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
            funds_receiving_stake_vkey=funds_receiving_stake_vkey,
            funds_receiving_stake_vkey_file=helpers._maybe_path(funds_receiving_stake_vkey_file),
            funds_receiving_stake_key_hash=funds_receiving_stake_key_hash,
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=helpers._maybe_path(deposit_return_stake_vkey_file),
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )
        return action_out

    def view(self, action_file: itp.FileType) -> dict[str, tp.Any]:
        """View a governance action vote."""
        action_file = pl.Path(action_file).expanduser()
        clusterlib_helpers._check_files_exist(action_file, clusterlib_obj=self._clusterlib_obj)

        stdout = self._clusterlib_obj.cli(
            [
                *self._group_args,
                "view",
                "--action-file",
                str(action_file),
            ]
        ).stdout.strip()
        stdout_dec = stdout.decode("utf-8") if stdout else ""

        out: dict[str, tp.Any] = json.loads(stdout_dec)
        return out

    def create_hardfork(
        self,
        action_name: str,
        deposit_amt: int,
        anchor_url: str,
        anchor_data_hash: str,
        protocol_major_version: int,
        protocol_minor_version: int,
        prev_action_txid: str = "",
        prev_action_ix: int = -1,
        deposit_return_stake_vkey: str = "",
        deposit_return_stake_vkey_file: itp.FileType | None = None,
        deposit_return_stake_key_hash: str = "",
        destination_dir: itp.FileType = ".",
    ) -> structs.ActionHardfork:
        """Create a hardfork initiation proposal."""
        destination_dir = pl.Path(destination_dir).expanduser()
        out_file = destination_dir / f"{action_name}_hardfork.action"
        clusterlib_helpers._check_files_exist(out_file, clusterlib_obj=self._clusterlib_obj)

        key_args = self._get_deposit_return_key_args(
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=deposit_return_stake_vkey_file,
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )

        anchor_args = self._get_anchor_args(
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
        )

        prev_action_args = self._get_optional_prev_action_args(
            prev_action_txid=prev_action_txid, prev_action_ix=prev_action_ix
        )

        self._clusterlib_obj.cli(
            [
                "cardano-cli",
                "conway",
                *self._group_args,
                "create-hardfork",
                *self.magic_args,
                "--governance-action-deposit",
                str(deposit_amt),
                *key_args,
                *prev_action_args,
                *anchor_args,
                "--protocol-major-version",
                str(protocol_major_version),
                "--protocol-minor-version",
                str(protocol_minor_version),
                "--out-file",
                str(out_file),
            ],
            add_default_args=False,
        )
        helpers._check_outfiles(out_file)

        action_out = structs.ActionHardfork(
            action_file=out_file,
            deposit_amt=deposit_amt,
            anchor_url=anchor_url,
            anchor_data_hash=anchor_data_hash,
            protocol_major_version=protocol_major_version,
            protocol_minor_version=protocol_minor_version,
            prev_action_txid=prev_action_txid,
            prev_action_ix=prev_action_ix,
            deposit_return_stake_vkey=deposit_return_stake_vkey,
            deposit_return_stake_vkey_file=helpers._maybe_path(deposit_return_stake_vkey_file),
            deposit_return_stake_key_hash=deposit_return_stake_key_hash,
        )
        return action_out



import itertools
import pathlib as pl
import random
import string

from cardano_clusterlib import exceptions
from cardano_clusterlib import types as itp


def get_rand_str(length: int = 8) -> str:
    """Return random ASCII lowercase string."""
    if length < 1:
        return ""
    return "".join(random.choice(string.ascii_lowercase) for i in range(length))


def read_address_from_file(addr_file: itp.FileType) -> str:
    """Read address stored in file."""
    with open(pl.Path(addr_file).expanduser(), encoding="utf-8") as in_file:
        return in_file.read().strip()


def _prepend_flag(flag: str, contents: itp.UnpackableSequence) -> list[str]:
    """Prepend flag to every item of the sequence.

    Args:
        flag: A flag to prepend to every item of the `contents`.
        contents: A list (iterable) of content to be prepended.

    Returns:
        List[str]: A list of flag followed by content, see below.

    >>> ClusterLib._prepend_flag(None, "--foo", [1, 2, 3])
    ['--foo', '1', '--foo', '2', '--foo', '3']
    """
    return list(itertools.chain.from_iterable([flag, str(x)] for x in contents))


def _check_outfiles(*out_files: itp.FileType) -> None:
    """Check that the expected output files were created.

    Args:
        *out_files: Variable length list of expected output files.
    """
    for out_file in out_files:
        out_file_p = pl.Path(out_file).expanduser()
        if not out_file_p.exists():
            msg = f"The expected file `{out_file}` doesn't exist."
            raise exceptions.CLIError(msg)


def _maybe_path(file: itp.FileType | None) -> pl.Path | None:
    """Return `Path` if `file` is thruthy."""
    return pl.Path(file) if file else None



"""Wrapper for cardano-cli for working with cardano cluster."""

import contextlib
import datetime
import functools
import json
import logging
import pathlib as pl
import typing as tp
import warnings

from cardano_clusterlib import clusterlib_helpers
from cardano_clusterlib import consts
from cardano_clusterlib import helpers
from cardano_clusterlib import structs
from cardano_clusterlib import txtools
from cardano_clusterlib import types as itp

LOGGER = logging.getLogger(__name__)


class QueryGroup:
    def __init__(self, clusterlib_obj: "itp.ClusterLib") -> None:
        self._clusterlib_obj = clusterlib_obj

    def _get_cred_args(
        self,
        drep_script_hash: str = "",
        drep_vkey: str = "",
        drep_vkey_file: itp.FileType | None = None,
        drep_key_hash: str = "",
    ) -> list[str]:
        """Get arguments for script or verification key."""
        if drep_script_hash:
            cred_args = ["--drep-script-hash", str(drep_script_hash)]
        elif drep_vkey:
            cred_args = ["--drep-verification-key", str(drep_vkey)]
        elif drep_vkey_file:
            cred_args = ["--drep-verification-key-file", str(drep_vkey_file)]
        elif drep_key_hash:
            cred_args = ["--drep-key-hash", str(drep_key_hash)]
        else:
            cred_args = []

        return cred_args

    def query_cli(
        self, cli_args: itp.UnpackableSequence, cli_sub_args: itp.UnpackableSequence = ()
    ) -> str:
        """Run the `cardano-cli query` command."""
        stdout = self._clusterlib_obj.cli(
            [
                "query",
                *cli_args,
                *self._clusterlib_obj.magic_args,
                *self._clusterlib_obj.socket_args,
                *cli_sub_args,
            ]
        ).stdout
        stdout_dec = stdout.decode("utf-8") if stdout else ""
        return stdout_dec

    def get_utxo(
        self,
        address: str | list[str] = "",
        txin: str | list[str] = "",
        utxo: structs.UTXOData | structs.OptionalUTXOData = (),
        tx_raw_output: structs.TxRawOutput | None = None,
        coins: itp.UnpackableSequence = (),
    ) -> list[structs.UTXOData]:
        """Return UTxO info for payment address.

        Args:
            address: Payment address(es).
            txin: Transaction input(s) (TxId#TxIx).
            utxo: Representation of UTxO data (`structs.UTXOData`).
            tx_raw_output: A data used when building a transaction (`structs.TxRawOutput`).
            coins: A list (iterable) of coin names (asset IDs, optional).

        Returns:
            List[structs.UTXOData]: A list of UTxO data.
        """
        cli_args = ["utxo", "--output-json"]

        address_single = ""
        sort_results = False
        if address:
            if isinstance(address, str):
                address_single = address
                address = [address]
            cli_args.extend(helpers._prepend_flag("--address", address))
        elif txin:
            if isinstance(txin, str):
                txin = [txin]
            cli_args.extend(helpers._prepend_flag("--tx-in", txin))
        elif utxo:
            if isinstance(utxo, structs.UTXOData):
                utxo = [utxo]
            utxo_formatted = [f"{u.utxo_hash}#{u.utxo_ix}" for u in utxo]
            cli_args.extend(helpers._prepend_flag("--tx-in", utxo_formatted))
        elif tx_raw_output:
            sort_results = True
            change_txout_num = 1 if tx_raw_output.change_address else 0
            return_collateral_txout_num = 1 if tx_raw_output.script_txins else 0
            num_of_txouts = (
                tx_raw_output.txouts_count + change_txout_num + return_collateral_txout_num
            )
            utxo_hash = self._clusterlib_obj.g_transaction.get_txid(
                tx_body_file=tx_raw_output.out_file
            )
            utxo_formatted = [f"{utxo_hash}#{ix}" for ix in range(num_of_txouts)]
            cli_args.extend(helpers._prepend_flag("--tx-in", utxo_formatted))
        else:
            msg = "Either `address`, `txin`, `utxo` or `tx_raw_output` need to be specified."
            raise AssertionError(msg)

        utxo_dict = json.loads(self.query_cli(cli_args))
        utxos = txtools.get_utxo(utxo_dict=utxo_dict, address=address_single, coins=coins)
        if sort_results:
            utxos = sorted(utxos, key=lambda u: u.utxo_ix)
        return utxos

    def get_tip(self) -> dict[str, tp.Any]:
        """Return current tip - last block successfully applied to the ledger."""
        tip: dict[str, tp.Any] = json.loads(self.query_cli(["tip"]))

        # "syncProgress" is returned as string
        sync_progress = tip.get("syncProgress")
        if sync_progress:
            with contextlib.suppress(ValueError):
                tip["syncProgress"] = float(sync_progress)

        return tip

    def get_ledger_state(self) -> dict:
        """Return the current ledger state info."""
        ledger_state: dict = json.loads(self.query_cli(["ledger-state"]))
        return ledger_state

    def save_ledger_state(
        self,
        state_name: str,
        destination_dir: itp.FileType = ".",
    ) -> pl.Path:
        """Save ledger state to file.

        Args:
            state_name: A name of the ledger state (can be epoch number, etc.).
            destination_dir: A path to directory for storing the state JSON file (optional).

        Returns:
            Path: A path to the generated state JSON file.
        """
        json_file = pl.Path(destination_dir) / f"{state_name}_ledger_state.json"
        # TODO: workaround for https://github.com/input-output-hk/cardano-node/issues/2461
        # self.query_cli(["ledger-state", "--out-file", str(json_file)])
        ledger_state = self.get_ledger_state()
        with open(json_file, "w", encoding="utf-8") as fp_out:
            json.dump(ledger_state, fp_out, indent=4)
        return json_file

    def get_protocol_state(self) -> dict:
        """Return the current protocol state info."""
        protocol_state: dict = json.loads(self.query_cli(["protocol-state"]))
        return protocol_state

    def get_protocol_params(self) -> dict:
        """Return the current protocol parameters."""
        self._clusterlib_obj.refresh_pparams_file()
        with open(self._clusterlib_obj.pparams_file, encoding="utf-8") as in_json:
            pparams: dict = json.load(in_json)
        return pparams

    def get_registered_stake_pools_ledger_state(self) -> dict:
        """Return ledger state info for registered stake pools."""
        registered_pools_details: dict = self.get_ledger_state()["stateBefore"]["esLState"][
            "delegationState"
        ]["pstate"]["pParams pState"]
        return registered_pools_details

    def get_stake_snapshot(
        self, stake_pool_ids: tp.Optional[list[str]] = None, all_stake_pools: bool = False
    ) -> dict[str, tp.Any]:
        """Return the three stake snapshots, plus the total active stake.

        Args:
            stake_pool_ids: A list of stake pool IDs, Bech32-encoded or hex-encoded (optional).
            all_stake_pools: A bool indicating whether to query for all stake pools (optional).

        Returns:
            dict: A stake snapshot data.
        """
        query_args = ["stake-snapshot"]

        if all_stake_pools:
            query_args.extend(["--all-stake-pools"])
        elif stake_pool_ids:
            query_args.extend(helpers._prepend_flag("--stake-pool-id", stake_pool_ids))

        stake_snapshot: dict[str, tp.Any] = json.loads(self.query_cli(query_args))
        return stake_snapshot

    def get_pool_params(
        self,
        stake_pool_id: str,
    ) -> structs.PoolParamsTop:
        """Return a pool parameters.

        Args:
            stake_pool_id: An ID of the stake pool (Bech32-encoded or hex-encoded).

        Returns:
            dict: A pool parameters.
        """
        warnings.warn(
            "`pool-params` deprecated by `pool-state` for node 1.35.4+",
            DeprecationWarning,
            stacklevel=2,
        )

        pool_params: dict = json.loads(
            self.query_cli(["pool-params", "--stake-pool-id", stake_pool_id])
        )

        # In node 1.35.1+ the information is nested under hex encoded stake pool ID
        if pool_params and "poolParams" not in pool_params:
            pool_params = next(iter(pool_params.values()))

        retiring = pool_params.get("retiring")  # pool retiring epoch
        pparams_top = structs.PoolParamsTop(
            pool_params=pool_params.get("poolParams") or {},
            future_pool_params=pool_params.get("futurePoolParams") or {},
            retiring=int(retiring) if retiring is not None else None,
        )
        return pparams_top

    def get_pool_state(
        self,
        stake_pool_id: str,
    ) -> structs.PoolParamsTop:
        """Return a pool state.

        Args:
            stake_pool_id: An ID of the stake pool (Bech32-encoded or hex-encoded).

        Returns:
            dict: A pool parameters.
        """
        pool_state: dict = json.loads(
            self.query_cli(["pool-state", "--stake-pool-id", stake_pool_id])
        )

        # The information is nested under hex encoded stake pool ID
        if pool_state:
            pool_state = next(iter(pool_state.values()))

        retiring = pool_state.get("retiring")  # pool retiring epoch
        pparams_top = structs.PoolParamsTop(
            pool_params=pool_state.get("poolParams") or {},
            future_pool_params=pool_state.get("futurePoolParams") or {},
            retiring=int(retiring) if retiring is not None else None,
        )
        return pparams_top

    def get_stake_addr_info(self, stake_addr: str) -> structs.StakeAddrInfo:
        """Return the current delegations and reward accounts filtered by stake address.

        Args:
            stake_addr: A stake address string.

        Returns:
            structs.StakeAddrInfo: A data container containing stake address info.
        """
        output_json = json.loads(self.query_cli(["stake-address-info", "--address", stake_addr]))
        if not output_json:
            return structs.StakeAddrInfo(
                address="",
                delegation="",
                reward_account_balance=0,
                registration_deposit=-1,
                vote_delegation="",
            )

        address_rec = next(iter(output_json))
        address = address_rec.get("address") or ""
        delegation = address_rec.get("delegation") or address_rec.get("stakeDelegation") or ""
        reward_account_balance = address_rec.get("rewardAccountBalance") or 0
        tmp_deposit = address_rec.get("stakeRegistrationDeposit") or address_rec.get(
            "delegationDeposit"
        )
        registration_deposit = -1 if tmp_deposit is None else tmp_deposit
        vote_delegation = address_rec.get("voteDelegation") or ""
        return structs.StakeAddrInfo(
            address=address,
            delegation=delegation,
            reward_account_balance=reward_account_balance,
            registration_deposit=registration_deposit,
            vote_delegation=vote_delegation,
        )

    def get_address_deposit(self, pparams: dict[str, tp.Any] | None = None) -> int:
        """Return stake address deposit amount."""
        pparams = pparams or self.get_protocol_params()
        return pparams.get("stakeAddressDeposit") or 0

    def get_pool_deposit(self, pparams: dict[str, tp.Any] | None = None) -> int:
        """Return stake pool deposit amount."""
        pparams = pparams or self.get_protocol_params()
        return pparams.get("stakePoolDeposit") or 0

    def get_drep_deposit(self, pparams: dict[str, tp.Any] | None = None) -> int:
        """Return DRep deposit amount."""
        pparams = pparams or self.get_protocol_params()
        return pparams.get("dRepDeposit") or 0

    def get_gov_action_deposit(self, pparams: dict[str, tp.Any] | None = None) -> int:
        """Return governance action deposit amount."""
        pparams = pparams or self.get_protocol_params()
        return pparams.get("govActionDeposit") or 0

    def get_stake_distribution(self) -> dict[str, float]:
        """Return current aggregated stake distribution per stake pool."""
        # Stake pool values are displayed starting with line 2 of the command output
        result = self.query_cli(["stake-distribution", "--output-text"]).splitlines()[2:]
        stake_distribution: dict[str, float] = {}
        for pool in result:
            pool_id, stake = pool.split()
            stake_distribution[pool_id] = float(stake)
        return stake_distribution

    def get_stake_pools(self) -> list[str]:
        """Return the node's current set of stake pool ids."""
        stake_pools = self.query_cli(["stake-pools", "--output-text"]).splitlines()
        return stake_pools

    def get_leadership_schedule(
        self,
        vrf_skey_file: itp.FileType,
        stake_pool_vkey: str = "",
        cold_vkey_file: itp.FileType | None = None,
        stake_pool_id: str = "",
        for_next: bool = False,
    ) -> list[structs.LeadershipSchedule]:
        """Get the slots the node is expected to mint a block in.

        Args:
            vrf_skey_file: A path to node VRF skey file.
            stake_pool_vkey: A pool cold vkey (Bech32 or hex-encoded, optional)
            cold_vkey_file: A path to pool cold vkey file (optional).
            stake_pool_id: An ID of the stake pool (Bech32 or hex-encoded, optional).
            for_next: A bool indicating whether to get the leadership schedule for the following
                epoch (current epoch by default)

        Returns:
            List[structs.LeadershipSchedule]: A list of `structs.LeadershipSchedule`, specifying
                slot and time.
        """
        args = []

        if stake_pool_vkey:
            args.extend(
                [
                    "--stake-pool-verification-key",
                    str(stake_pool_vkey),
                ]
            )
        elif cold_vkey_file:
            args.extend(
                [
                    "--cold-verification-key-file",
                    str(cold_vkey_file),
                ]
            )
        elif stake_pool_id:
            args.extend(
                [
                    "--stake-pool-id",
                    str(stake_pool_id),
                ]
            )
        else:
            msg = "Either `stake_pool_vkey`, `cold_vkey_file` or `stake_pool_id` is needed."
            raise AssertionError(msg)

        args.append("--next" if for_next else "--current")

        unparsed = self.query_cli(
            [
                "leadership-schedule",
                "--output-text",
                "--genesis",
                str(self._clusterlib_obj.genesis_json),
                "--vrf-signing-key-file",
                str(vrf_skey_file),
                *args,
            ]
            # Schedule values are displayed starting with line 2 of the command output
        ).splitlines()[2:]

        schedule = []
        for rec in unparsed:
            slot_no, date_str, time_str, *__ = rec.split()
            # Add milliseconds component of a time string if it is missing
            time_str = time_str if "." in time_str else f"{time_str}.0"
            schedule.append(
                structs.LeadershipSchedule(
                    slot_no=int(slot_no),
                    utc_time=datetime.datetime.strptime(
                        f"{date_str} {time_str}", "%Y-%m-%d %H:%M:%S.%f"
                    ).replace(tzinfo=datetime.timezone.utc),
                )
            )

        return schedule

    def get_slot_no(self, tip: dict[str, tp.Any] | None = None) -> int:
        """Return slot number of last block that was successfully applied to the ledger."""
        tip = tip or self.get_tip()
        return int(self.get_tip()["slot"])

    def get_block_no(self, tip: dict[str, tp.Any] | None = None) -> int:
        """Return block number of last block that was successfully applied to the ledger."""
        tip = tip or self.get_tip()
        return int(self.get_tip()["block"])

    def get_epoch(self, tip: dict[str, tp.Any] | None = None) -> int:
        """Return epoch of last block that was successfully applied to the ledger."""
        tip = tip or self.get_tip()
        return int(self.get_tip()["epoch"])

    def get_epoch_slot_no(self, tip: dict[str, tp.Any] | None = None) -> int:
        """Return slot number within a given epoch.

        (of last block successfully applied to the ledger)
        """
        tip = tip or self.get_tip()
        return int(self.get_tip()["slotInEpoch"])

    def get_slots_to_epoch_end(self, tip: dict[str, tp.Any] | None = None) -> int:
        """Return the number of slots left until the epoch end."""
        tip = tip or self.get_tip()
        return int(self.get_tip()["slotsToEpochEnd"])

    def get_era(self, tip: dict[str, tp.Any] | None = None) -> str:
        """Return network era."""
        tip = tip or self.get_tip()
        era: str = self.get_tip()["era"]
        return era

    def get_address_balance(self, address: str, coin: str = consts.DEFAULT_COIN) -> int:
        """Get total balance of an address (sum of all UTxO balances).

        Args:
            address: A payment address string.
            coin: A coin name (asset IDs).

        Returns:
            int: A total balance.
        """
        utxo = self.get_utxo(address=address, coins=[coin])
        address_balance = functools.reduce(lambda x, y: x + y.amount, utxo, 0)
        return int(address_balance)

    def get_utxo_with_highest_amount(
        self, address: str, coin: str = consts.DEFAULT_COIN
    ) -> structs.UTXOData:
        """Return data for UTxO with the highest amount.

        Args:
            address: A payment address string.
            coin: A coin name (asset IDs).

        Returns:
            structs.UTXOData: A UTxO record with the highest amount.
        """
        utxo = self.get_utxo(address=address, coins=[coin])
        highest_amount_rec = max(utxo, key=lambda x: x.amount)
        return highest_amount_rec

    def get_kes_period(self) -> int:
        """Return last block KES period."""
        return int(self.get_slot_no() // self._clusterlib_obj.slots_per_kes_period)

    def get_kes_period_info(self, opcert_file: itp.FileType) -> dict[str, tp.Any]:
        """Get information about the current KES period and node's operational certificate.

        Args:
            opcert_file: A path to operational certificate.

        Returns:
            dict: A dictionary containing KES period information.
        """
        command_output = self.query_cli(["kes-period-info", "--op-cert-file", str(opcert_file)])
        return clusterlib_helpers._get_kes_period_info(kes_info=command_output)

    def get_mempool_info(self) -> dict[str, tp.Any]:
        """Return info about the current mempool's capacity and sizes.

        Returns:
            dict: A dictionary containing mempool information.
        """
        tx_mempool: dict[str, tp.Any] = json.loads(
            self.query_cli(["tx-mempool"], cli_sub_args=[consts.SUBCOMMAND_MARK, "info"])
        )
        return tx_mempool

    def get_mempool_next_tx(self) -> dict[str, tp.Any]:
        """Return info about the next transaction in the mempool's current list.

        Returns:
            dict: A dictionary containing mempool information.
        """
        tx_mempool: dict[str, tp.Any] = json.loads(
            self.query_cli(["tx-mempool"], cli_sub_args=[consts.SUBCOMMAND_MARK, "next-tx"])
        )
        return tx_mempool

    def get_mempool_tx_exists(self, txid: str) -> dict[str, tp.Any]:
        """Query if a particular transaction exists in the mempool.

        Args:
            txid: A transaction ID.

        Returns:
            dict: A dictionary containing mempool information.
        """
        tx_mempool: dict[str, tp.Any] = json.loads(
            self.query_cli(["tx-mempool"], cli_sub_args=[consts.SUBCOMMAND_MARK, "tx-exists", txid])
        )
        return tx_mempool

    def get_slot_number(self, timestamp: datetime.datetime) -> int:
        """Return slot number for UTC timestamp."""
        timestamp_str = timestamp.strftime("%Y-%m-%dT%H:%M:%SZ")

        slot_number: int = json.loads(self.query_cli(["slot-number", timestamp_str]))

        return slot_number

    def get_constitution(self) -> dict[str, tp.Any]:
        """Get the constitution."""
        out: dict[str, tp.Any] = json.loads(self.query_cli(["constitution"]))
        return out

    def get_gov_state(self) -> dict[str, tp.Any]:
        """Get the governance state."""
        out: dict[str, tp.Any] = json.loads(self.query_cli(["gov-state"]))
        return out

    def get_drep_state(
        self,
        drep_script_hash: str = "",
        drep_vkey: str = "",
        drep_vkey_file: itp.FileType | None = None,
        drep_key_hash: str = "",
    ) -> list[list[dict[str, tp.Any]]]:
        """Get the DRep state.

        When no key is provided, query all DReps.

        Args:
            drep_script_hash: DRep script hash (hex-encoded, optional).
            drep_vkey: DRep verification key (Bech32 or hex-encoded).
            drep_vkey_file: Filepath of the DRep verification key.
            drep_key_hash: DRep verification key hash (either Bech32-encoded or hex-encoded).

        Returns:
            list[list[dict[str, Any]]]: DRep state.
        """
        cred_args = self._get_cred_args(
            drep_script_hash=drep_script_hash,
            drep_vkey=drep_vkey,
            drep_vkey_file=drep_vkey_file,
            drep_key_hash=drep_key_hash,
        )
        if not cred_args:
            cred_args = ["--all-dreps"]

        out: list[list[dict[str, tp.Any]]] = json.loads(self.query_cli(["drep-state", *cred_args]))
        return out

    def get_drep_stake_distribution(
        self,
        drep_script_hash: str = "",
        drep_vkey: str = "",
        drep_vkey_file: itp.FileType | None = None,
        drep_key_hash: str = "",
    ) -> dict[str, tp.Any]:
        """Get the DRep stake distribution.

        When no key is provided, query all DReps.

        Args:
            drep_script_hash: DRep script hash (hex-encoded, optional).
            drep_vkey: DRep verification key (Bech32 or hex-encoded).
            drep_vkey_file: Filepath of the DRep verification key.
            drep_key_hash: DRep verification key hash (either Bech32-encoded or hex-encoded).

        Returns:
            dict[str, Any]: DRep stake distribution.
        """
        cred_args = self._get_cred_args(
            drep_script_hash=drep_script_hash,
            drep_vkey=drep_vkey,
            drep_vkey_file=drep_vkey_file,
            drep_key_hash=drep_key_hash,
        )
        if not cred_args:
            cred_args = ["--all-dreps"]

        out: list[list] | dict[str, tp.Any] = json.loads(
            self.query_cli(["drep-stake-distribution", *cred_args])
        )
        recs: dict[str, tp.Any] = {i[0]: i[1] for i in out} if isinstance(out, list) else out
        return recs

    def get_committee_state(self) -> dict[str, tp.Any]:
        """Get the committee state."""
        out: dict[str, tp.Any] = json.loads(self.query_cli(["committee-state"]))
        return out

    def get_treasury(self) -> int:
        """Get the treasury value."""
        return int(self.query_cli(["treasury"]))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: clusterlib_obj={id(self._clusterlib_obj)}>"
